import argparse

from . import Command, register
from .. import models, auth
from ..models import User


@register
class AddUser(Command):
    NAME = "adduser"
    HELP = "Add a user to the database."
    DESC = "Adds a user to the database with the prefered roles."


    def _initArgParser(self, parser):
        parser.add_argument("-l", "--list-roles", dest="list_roles",
                            action="store_true", help="list the roles available")
        parser.add_argument("username", nargs="?", help="users name")
        parser.add_argument("password", nargs="?", help="users password")
        parser.add_argument("roles", nargs=argparse.REMAINDER,
                            help="users roles")


    def _run(self, args=None):
        super()._run()
        args = args or self.args

        if args.list_roles:
            print("Subsonic API roles: [" +
                  ", ".join(auth.Roles.subsonic_roles) + "]")
            print("Admin user roles: [" +
                  ", ".join(auth.Roles.admin_roles) + "]")
            print("Default user roles: [" +
                  ", ".join(auth.Roles.def_user_roles) + "]")
            return 0

        if not args.username or not args.password:
            print("Missing username or password arguments.")
            return -1

        if len(args.roles):
            for role in [auth.Roles.REST, auth.Roles.USERS]:
                if role not in args.roles:
                    args.roles.append(role)
        else:
            args.roles = auth.Roles.def_user_roles

        res = self.db_session.query(User).filter(User.name == args.username)
        if res.one_or_none():
            print("User '%s' already exists." % args.username)
            return -1
        else:
            ret = models.addUser(self.db_session, args.username,
                                 args.password, args.roles)
            if ret:
                print("Added user '%s'." % args.username)
                return 0
            else:
                print(ret)
                return -1



from . import Command, registerCmd, NotFound, album_t, fillAlbumID3
from ...models import Album


@registerCmd
class GetAlbum(Command):
    name = "getAlbum.view"
    param_defs = {"id": {"required": True, "type": album_t}}
    dbsess = True


    def handleReq(self, session):
        album = None
        for row in session.query(Album).filter(
                Album.id == self.params["id"]).all():
            album = fillAlbumID3(session, row, self.req.authed_user, True)
        if album is None:
            raise NotFound(self.req.params["id"])
        return self.makeResp(child=album)



import xml.etree.ElementTree as ET

from . import (Command, registerCmd, MissingParam, NotFound, fillAlbumUser,
               fillTrackUser)
from ...models import Artist, Album, Track


@registerCmd
class GetMusicDirectory(Command):
    name = "getMusicDirectory.view"
    param_defs = {"id": {"required": True}}
    dbsess = True


    def __init__(self, route, req, session=None):
        super(GetMusicDirectory, self).__init__(route, req, session)
        self.setParams()


    def setParams(self, dir_param="directory", album_param="child",
                  track_param="child"):
        self.dir_param = dir_param
        self.album_param = album_param
        self.track_param = track_param


    def handleReq(self, session):
        directory = ET.Element(self.dir_param)
        directory.set("id", self.params["id"])
        # FIXME: Do we care about the top level directory hierarchy?
        if self.params["id"].startswith("fl-"):
            raise Exception("TOP LEVEL FOLDER")
        elif self.params["id"].startswith("ar-"):
            artist_id = int(self.params["id"][3:])
            row = session.query(Artist).filter(
                Artist.id == artist_id).one_or_none()
            if row:
                directory.set("name", row.name)
                directory.set("parent", "fl-%s" % row.lib_id)
            else:
                raise NotFound(self.params["id"])
            # Gather albums
            for row in session.query(Album).filter(
                    Album.artist_id == artist_id).all():
                album = fillAlbumUser(session, row, None, self.req.authed_user,
                                      self.album_param)
                directory.append(album)
            # Gather tracks with no album
            for row in session.query(Track).filter(
                    Track.album_id is None, Track.artist_id == artist_id).\
                    order_by(Track.track_num).all():
                song = fillTrackUser(session, row, None, self.req.authed_user,
                                     self.track_param)
                directory.append(song)
        elif self.params["id"].startswith("al-"):
            album_id = int(self.params["id"][3:])
            album = session.query(Album).filter(
                Album.id == album_id).one_or_none()
            if not album:
                raise NotFound(self.params["id"])
            directory.set("parent", "ar-%d" % album.artist_id)
            directory.set("name", album.title)
            directory.set("id", self.params["id"])
            song = None
            for row in session.query(Track).filter(
                    Track.album_id == album_id).order_by(Track.track_num).all():
                song = fillTrackUser(session, row, None, self.req.authed_user,
                                     self.track_param)
                directory.append(song)
            if song is None:
                raise NotFound(self.params["id"])
        elif self.params["id"].startswith("tr-"):
            track_id = int(self.params["id"][3:])
            row = session.query(Track).filter(Track.id == track_id).one()
            if row is None:
                raise NotFound(self.params["id"])
            song = fillTrackUser(session, row, None, self.req.authed_user,
                                 self.track_param)
            if row.album:
                directory.set("parent", "al-%d" % row.album.id)
                directory.set("name", row.album.title)
            elif row.artist:
                directory.set("parent", "ar-%d" % row.artist.id)
                directory.set("name", row.artist.name)
            if self.dir_param == "song":
                directory = song
            else:
                directory.append(song)
        else:
            raise MissingParam("Invalid value for 'id'")
        return self.makeResp(child=directory)



from . getalbumlist import GetAlbumList
from . import registerCmd, fillAlbumID3


@registerCmd
class GetAlbumList2(GetAlbumList):
    name = "getAlbumList2.view"
    param_defs = GetAlbumList.param_defs
    dbsess = True


    def __init__(self, route, req):
        super().__init__(route, req)
        self.setParams("albumList2")


    def processRows(self, session, alist, result):
        for row in result:
            album = fillAlbumID3(session, row, self.req.authed_user, False)
            alist.append(album)



from pyramid.response import FileResponse

from . import Command, registerCmd, bool_t, track_t, int_t
from ...models import Track


@registerCmd
class Stream(Command):
    name = "stream.view"
    param_defs = {
        "id": {"required": True, "type": track_t},
        "maxBitRate": {"type": int_t},
        "format": {"values": ["mp3", "flv", "raw"]},
        "timeOffset": {"type": int_t},
        "size": {"type": int_t},
        "estimateContentLength": {"default": False, "type": bool_t},
        }
    dbsess = True


    def handleReq(self, session):
        row = session.query(Track).filter(Track.id == self.params["id"]).all()[0]
        return FileResponse(row.path)



from .getalbuminfo import GetAlbumInfo
from . import registerCmd


# TODO: Actually implement
@registerCmd
class GetAlbumInfo2(GetAlbumInfo):
    name = "getAlbumInfo2.view"
    param_defs = GetAlbumInfo.param_defs
    dbsess = True



from . import Command, registerCmd, NotFound
from ...models import Image, getUserByName


@registerCmd
class GetAvatar(Command):
    name = "getAvatar.view"
    param_defs = {
        "username": {"required": True},
        }
    dbsess = True


    def handleReq(self, session):
        if self.req.authed_user.name == self.params["username"]:
            db_user = self.req.authed_user
        else:
            db_user = getUserByName(session, self.params["username"])
        if not db_user:
            raise NotFound("User not found")
        row = session.query(Image).\
                  filter(Image.id == db_user.avatar).one_or_none()
        if not row:
            raise NotFound("User has no avatar")
        return self.makeBinaryResp(row.data, row.mime_type, row.md5)



import time
import threading
from multiprocessing import Process
from sqlalchemy import func

from . import __main__, web
from .log import log
from .models import Track


sync_proc = None
last_sync_ret = None
check_thread = None


def doSync():
    log.info("Starting music sync...")
    __main__.main(["-c", web.CONFIG_FILE, "sync"])
    return True


def startSync(session):
    global sync_proc
    log.info("Starting music sync subprocess...")
    startChecker()
    sync_proc = Process(target=doSync, name="Sync Task")
    sync_proc.start()
    return syncStatus(session)


def syncStatus(session):
    count = session.query(func.count(Track.id)).one()[0]
    if sync_proc and sync_proc.is_alive():
        return (True, count)
    else:
        return (False, count)


def startChecker():
    global check_thread
    if check_thread:
        return
    check_thread = threading.Thread(target=checkProc, name="Sync proc checker")
    check_thread.daemon = True
    check_thread.start()


def checkProc():
    global sync_proc
    while True:
        time.sleep(5.0)
        if sync_proc and not sync_proc.is_alive():
            log.debug("Joining sync proccess...")
            sync_proc.join()
            log.debug("Joined sync proccess")
            sync_proc = None



from pyramid.response import Response
from pyramid.exceptions import NotFound

from . import Command, registerCmd, MissingParam
from ...models import Image


@registerCmd
class GetCoverArt(Command):
    name = "getCoverArt.view"
    param_defs = {"id": {"required": True}}
    dbsess = True


    def handleReq(self, session):
        id = self.params["id"]
        try:
            num = int(id[3:])
        except:
            raise MissingParam("Invalid id: %s" % id)
        if id.startswith("ar-") or id.startswith("al-"):
            image = session.query(Image).filter_by(id=num).all()

        if len(image) == 1:
            return Response(content_type=image[0].mime_type,
                            body=image[0].data)
        else:
            raise NotFound()



import xml.etree.ElementTree as ET

from ... import sync
from . import Command, registerCmd


@registerCmd
class StartScan(Command):
    name = "startScan.view"
    param_defs = {}
    dbsess = True


    def handleReq(self, session):
        status = ET.Element("scanStatus")
        scanning, count = sync.startSync(session)
        status.set("scanning", "true" if scanning else "false")
        status.set("count", str(count))
        return self.makeResp(child=status)



from datetime import datetime
import xml.etree.ElementTree as ET

from . import Command, registerCmd, fillArtistUser, folder_t, datetime_t
from ...models import Artist, Meta


@registerCmd
class GetIndexes(Command):
    name = "getIndexes.view"
    param_defs = {
        "musicFolderId": {"type": folder_t},
        "ifModifiedSince": {"type": datetime_t},
        }
    dbsess = True


    def handleReq(self, session):
        indexes = ET.Element("indexes")
        index_group = None
        row = session.query(Meta).one_or_none()

        # Return empty response if sync is older
        if (self.params["ifModifiedSince"] and
                (self.params["ifModifiedSince"] >=
                 datetime.utcfromtimestamp(row.last_sync.timestamp()))):
            return self.makeResp()

        # TODO: Use libraries.last_sync once its done in mishmash
        indexes.set("lastModified",
                    str(int(row.last_sync.timestamp() * 1000)))
        # TODO: find the actual ignored articles
        indexes.set("ignoredArticles", "")

        q = session.query(Artist)
        if self.params["musicFolderId"]:
            q = q.filter(Artist.lib_id == self.params["musicFolderId"])
        rows = q.order_by(Artist.sort_name).all()
        for row in rows:
            first = row.sort_name[0].upper()
            if index_group != first:
                index_group = first
                index = ET.Element("index")
                indexes.append(index)
                index.set("name", index_group)
            artist = fillArtistUser(session, row, None, self.req.authed_user)
            index.append(artist)
        return self.makeResp(child=indexes)



from .getartistinfo import GetArtistInfo
from . import registerCmd


# TODO: Actually implement
@registerCmd
class GetArtistInfo2(GetArtistInfo):
    name = "getArtistInfo2.view"
    param_defs = GetArtistInfo.param_defs
    dbsess = True



from . import Command, register


@register
class Config(Command):
    NAME = "config"
    HELP = "See/edit the Unsonic configuration."
    DESC = "A general interface for various configurable toggles."


    def _initArgParser(self, parser):
        parser.add_argument("-l", "--list", dest="list", action="store_true",
                            help="list the configurables available")
        parser.add_argument("-s", "--set", dest="set", metavar="KEY",
                            help="set the named configurable to VALUE")
        parser.add_argument("-g", "--get", dest="list", metavar="KEY",
                            help="get the named configurable value")
        parser.add_argument("VALUE", nargs="?",
                            help="value of the configurable")


    def _run(self, args=None):
        super()._run()

        args = args or self.args



from . import registerCmd, track_t
from .stream import Stream


@registerCmd
class Download(Stream):
    name = "download.view"
    param_defs = {
        "id": {"required": True, "type": track_t},
        }



import xml.etree.ElementTree as ET

from . import BACON_IPSUM, Command, registerCmd, playable_id_t


# TODO: Actually implement
@registerCmd
class GetAlbumInfo(Command):
    name = "getAlbumInfo.view"
    param_defs = {
        "id": {"required": True, "type": playable_id_t},
        }
    dbsess = True


    # Actually do this for realz once last.fm stuff is hooked up
    def handleReq(self, session):
        ainfo = ET.Element("albumInfo")
        notes = ET.Element("notes")
        notes.text = BACON_IPSUM
        ainfo.append(notes)
        mbid = ET.Element("musicBrainzId")
        mbid.text = "1234567890"
        ainfo.append(mbid)

        for name in ["lastFmUrl", "smallImageUrl", "mediumImageUrl",
                     "largeImageUrl"]:
            url = ET.Element(name)
            url.text = "https://foo.com/foo"
            ainfo.append(url)

        return self.makeResp(child=ainfo)



import os
import json
import xmltodict
import logging
from collections import OrderedDict
from datetime import datetime
import xml.etree.ElementTree as ET

from pyramid.security import Allow, Authenticated, DENY_ALL

from eyed3.core import Date as Eyed3Date
from nicfit.console.ansi import Fg

from ...log import log
from ...version import PROTOCOL_VERSION, UNSONIC_PROTOCOL_VERSION
from ...models import Session, ArtistRating, AlbumRating, TrackRating, Track
from ...auth import Roles


BACON_IPSUM = (
"Bacon ipsum dolor amet biltong sausage ribeye pancetta salami pork chop. Short "
"loin sirloin burgdoggen, turducken kielbasa corned beef landjaeger chicken "
"short ribs capicola. Drumstick turkey jerky, cow shankle flank pork loin ball "
"tip. Meatball shoulder landjaeger jerky. Bresaola prosciutto alcatra venison, "
"meatloaf pork belly ball tip tail cupim porchetta. Chuck alcatra leberkas tail "
"flank. Kevin chicken strip steak meatball ground round cow.")


XML_HEADER = '<?xml version="1.0" encoding="UTF-8"?>'

commands = {}


class RouteContext(object):
    __acl__ = [(Allow, Authenticated, Roles.REST), DENY_ALL]

    def __init__(self, request):
        pass


class NoPerm(Exception):
    pass


class MissingParam(Exception):
    pass


class NotFound(Exception):
    pass


class InternalError(Exception):
    pass


class Command(object):
    E_GENERIC = ("0", "An unknown error occured")
    E_MISSING_PARAM = ("10", "Missing a required parameter")
    E_VER_CLIENT = ("20", "Incompatible Subsonic REST protocol version. " +
                          "Client must upgrade.")
    E_VER_SERVER = ("30", "Incompatible Subsonic REST protocol version. " +
                          "Server must upgrade.")
    E_AUTH = ("40", "Username or password incorrect")
    E_PERM = ("50", "Permission denied for this operation")
    # 60, trial period over, intentionally skipped, cause screw that noise.
    E_NOT_FOUND = ("70", "Requsted data not found")


    def __init__(self, route, req, session=None):
        self.req = req
        self.route = route
        self.params = {}
        # For testing
        self.session = session


    def __call__(self):
        try:
            self.parseParams()
            if hasattr(self, "dbsess"):
                if self.session:
                    return self.handleReq(self.session)
                else:
                    with Session() as session:
                        return self.handleReq(session)
            else:
                return self.handleReq()
        except MissingParam as e:
            return self.makeResp(status=(Command.E_MISSING_PARAM, str(e)))
        except NotFound as e:
            return self.makeResp(status=(Command.E_NOT_FOUND, str(e)))
        except InternalError as e:
            return self.makeResp(status=(Command.E_GENERIC, str(e)))
        except NoPerm as e:
            return self.makeResp(status=(Command.E_PERM, str(e)))


    def handleReq(self, session=None):
        raise Exception("Command must implement handleReq()")


    def makeBody(self, attrs, child, status):
        body = ET.Element("subsonic-response")
        attrs_ = {"status": "ok" if status is True else "failed",
                  "xmlns": "http://subsonic.org/restapi",
                  "version": PROTOCOL_VERSION,
                  "unsonic": UNSONIC_PROTOCOL_VERSION}
        attrs_.update(attrs)
        for key, value in attrs_.items():
            body.set(key, value)
        if status is not True and status is not False:
            error = ET.Element("error")
            if isinstance(status[0], tuple):
                error.set("code", status[0][0])
                error.set("message", "%s: %s" % (status[0][1], status[1]))
            else:
                error.set("code", status[0])
                error.set("message", status[1])
            body.append(error)
        if child is not None:
            body.append(child)
        return "%s%s\n" % (XML_HEADER, ET.tostring(body).decode("utf-8"))


    def toDict(self, body):
        root = xmltodict.parse(body, attr_prefix="")

        def walker(d):
            if not isinstance(d, list) and not isinstance(d, OrderedDict):
                return
            for key, val in d.items():
                if isinstance(val, list):
                    for val2 in val:
                        walker(val2)
                elif isinstance(val, OrderedDict):
                    walker(val)
                elif val is None:
                    d[key] = OrderedDict()
                elif val == "false":
                    d[key] = False
                elif val == "true":
                    d[key] = True

        walker(root)
        return root


    def makeResp(self, attrs={}, child=None, status=True, body=None):
        if body is None:
            body = self.makeBody(attrs, child, status)
        elif isinstance(body, ET.Element):
            body = "%s%s" % (XML_HEADER, ET.tostring(body).decode("utf-8"))
        resp = self.req.response
        pretty = None
        if "f" in self.req.params:
            if self.req.params["f"] == "jsonp" and "callback" in self.req.params:
                dct = self.toDict(body)
                if log.isEnabledFor(logging.DEBUG):
                    pretty = "%s(%s)" % (self.req.params["callback"],
                                         json.dumps(dct, indent=3))
                else:
                    pretty = body
                txt = "%s(%s)" % (self.req.params["callback"], json.dumps(dct))
                resp.text = txt
                resp.content_type = "application/javascript"
            elif self.req.params["f"] == "json":
                dct = self.toDict(body)
                if log.isEnabledFor(logging.DEBUG):
                    pretty = json.dumps(dct, indent=3)
                else:
                    pretty = body
                resp.text = json.dumps(dct)
                resp.content_type = "application/json"
        if not pretty:
            pretty = body
            resp.text = body
            resp.content_type = "text/xml"
        resp.charset = "UTF-8"
        log.debug("Response(%s): %s" % (self.name, Fg.blue(pretty)))
        return resp


    def makeBinaryResp(self, binary, mimetype, md5=None):
        resp = self.req.response
        resp.content_type = mimetype
        if md5:
            resp.content_md5 = md5
        resp.body = binary
        return resp


    def parseParams(self):
        mparams = self.req.params.mixed()
        for name, values in self.param_defs.items():
            if name in mparams:
                val = mparams[name]
                if "type" in values:
                    if "multi" in values:
                        if not isinstance(val, list):
                            val = [val]
                        lval = []
                        for v in val:
                            lval.append(values["type"](v))
                        val = lval
                    else:
                        val = values["type"](val)
                self.params[name] = val
                if "values" in values and val not in values["values"]:
                    raise MissingParam("Invalid type for param: %s" % name)
            else:
                if "default" in values:
                    self.params[name] = values["default"]
                else:
                    if "multi" in values:
                        self.params[name] = []
                    else:
                        self.params[name] = None
                if "required" in values and values["required"]:
                    raise MissingParam(name)


def registerCmd(cmd):
    commands[cmd.name] = cmd
    return cmd


# Param type check functions
def str_t(value):
    if not value or not len(value):
        raise MissingParam("Missing string parameter")
    return value


def int_t(value):
    try:
        return int(value)
    except:
        raise MissingParam("Invalid number parameter")


def bool_t(value):
    if value in ["True", "true"]:
        return True
    elif value in ["False", "false"]:
        return False
    else:
        raise MissingParam("Invalid type")


def positive_t(value):
    val = int(value)
    if val < 0:
        raise MissingParam("Invalid number, can not be negative")
    else:
        return val


def playable_id_t(value):
    for prefix in ["ar-", "al-", "tr-"]:
        if value.startswith(prefix):
            break
    else:
        raise MissingParam("Invalid id")
    return value


def artist_t(value):
    if not value.startswith("ar-"):
        raise MissingParam("Invalid id")
    return int(value[3:])


def album_t(value):
    if not value.startswith("al-"):
        raise MissingParam("Invalid id")
    return int(value[3:])


def track_t(value):
    if not value.startswith("tr-"):
        raise MissingParam("Invalid id")
    return int(value[3:])


def playlist_t(value):
    if not value.startswith("pl-"):
        raise MissingParam("Invalid id")
    return int(value[3:])


def folder_t(value):
    if not value.startswith("fl-"):
        raise MissingParam("Invalid id")
    return int(value[3:])


def datetime_t(tstamp):
    try:
        return datetime.utcfromtimestamp(int(tstamp) / 1000)
    except:
        raise MissingParam(
            "Invalid type for param. '%s' is not a timestamp" % tstamp)


def year_t(year):
    try:
        return Eyed3Date(int(year), 1, 1)
    except:
        raise MissingParam("Invalid type for param. '%s' is not a year" % year)


def bitrate_t(value):
    try:
        i = int(value)
    except:
        raise MissingParam("Invalid type for param. '%s' is not a number" %
                           value)
    if i in [0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320]:
        return i
    else:
        raise MissingParam("Invalid value for param. '%s' is not allowed" %
                           value)


def strDate(d):
    if isinstance(d, Eyed3Date):
        return datetime(d.year if d.year else 0,
                        d.month if d.month else 0,
                        d.day if d.day else 0,
                        d.hour if d.hour else 0,
                        d.minute if d.minute else 0,
                        d.second if d.second else 0).isoformat()
    else:
        return d.isoformat()


# Utilities for wrangling data into xml form
def fillCoverArt(session, row, elem, name):
    if row.images is not None and len(row.images) > 0:
        elem.set("coverArt", "%s-%d" % (name, row.images[0].id))
        for art in row.images:
            sub = ET.Element("cover-art")
            sub.text = "%s-%d" % (name, art.id)
            elem.append(sub)


def fillArtist(session, row, name="artist"):
    artist = ET.Element(name)
    artist.set("id", "ar-%d" % row.id)
    artist.set("name", row.name)
    fillCoverArt(session, row, artist, "ar")
    return artist


def fillArtistUser(session, artist_row, rating_row, user, name="artist"):
    artist = fillArtist(session, artist_row, name=name)
    if not rating_row:
        rating_row = session.query(ArtistRating).\
                         filter(ArtistRating.artist_id == artist_row.id,
                                ArtistRating.user_id == user.id).\
                         one_or_none()
    if rating_row and rating_row.starred:
        artist.set("starred", rating_row.starred.isoformat())
    return artist


def fillAlbum(session, row, name="album"):
    album = ET.Element(name)
    album.set("id", "al-%d" % row.id)
    album.set("album", row.title)
    album.set("title", row.title)
    album.set("isDir", "true")
    if row.artist:
        album.set("parent", "ar-%s" % row.artist.id)
    else:
        album.set("parent", "UNKNOWN")
    fillCoverArt(session, row, album, "al")
    if row.date_added:
        album.set("created", strDate(row.date_added))
    if row.artist and row.artist.name:
        album.set("artist", row.artist.name)
        album.set("artistId", "ar-%d" % row.artist.id)
    return album


def fillAlbumUser(session, album_row, rating_row, user, name="album"):
    album = fillAlbum(session, album_row, name=name)
    if not rating_row:
        rating_row = session.query(AlbumRating).\
                         filter(AlbumRating.album_id == album_row.id,
                                AlbumRating.user_id == user.id).one_or_none()
    if rating_row and rating_row.starred and not rating_row.pseudo_starred:
        album.set("starred", rating_row.starred.isoformat())
    return album


def fillAlbumID3(session, row, user, append_tracks):
    album = ET.Element("album")
    album.set("id", "al-%d" % row.id)
    album.set("name", row.title)
    fillCoverArt(session, row, album, "al")
    if row.date_added:
        album.set("created", strDate(row.date_added))
    if row.artist and row.artist.name:
        album.set("artist", row.artist.name)
        album.set("artistId", "ar-%d" % row.artist.id)
    if row.getBestDate():
        album.set("year", str(row.getBestDate().year))
    # TODO: what if more than one tag/genre? What to prefer?
    if len(row.tags):
        album.set("genre", row.tags[0].name)
    track_count = 0
    duration = 0
    for row in session.query(Track).filter(Track.album_id == row.id).all():
        track_count += 1
        duration = duration + row.time_secs
        if append_tracks:
            track = fillTrackUser(session, row, None, user)
            album.append(track)
    album.set("songCount", str(track_count))
    album.set("duration", str(duration))
    return album


def fillTrack(session, row, name="song"):
    song = ET.Element(name)
    song.set("id", "tr-%d" % row.id)
    if row.album_id:
        song.set("parent", "al-%d" % row.album_id)
    else:
        song.set("parent", "UNKNOWN")
    song.set("title", row.title)
    song.set("isDir", "false")
    album_name = "-"
    if row.album and row.album.title:
        album_name = row.album.title
    song.set("album", album_name)
    artist_name = "-"
    if row.artist and row.artist.name:
        artist_name = row.artist.name
    song.set("artist", artist_name)
    if row.track_num:
        song.set("track", str(row.track_num))
    if row.album and row.album.getBestDate():
        song.set("year", str(row.album.getBestDate().year))
    # TODO: what if more than one tag/genre? What to prefer?
    if len(row.tags):
        song.set("genre", row.tags[0].name)
    if row.album is not None:
        fillCoverArt(session, row.album, song, "al")
    song.set("size", str(row.size_bytes))
    # FIXME
    song.set("contentType", "audio/mpeg")
    song.set("transcodedContentType", "audio/mpeg")
    suffix = os.path.basename(row.path).split(".")
    suffix = suffix[-1] if len(suffix) else None
    if not suffix:
        suffix = "mp3"
    song.set("suffix", suffix)
    song.set("transcodedSuffix", suffix)
    song.set("duration", str(row.time_secs))
    song.set("bitRate", str(row.bit_rate))
    song.set("path", os.path.join(artist_name, album_name, row.title))
    song.set("isVideo", "false")
    return song


def fillTrackUser(session, song_row, rating_row, user, name="song"):
    song = fillTrack(session, song_row, name=name)
    if not rating_row:
        rating_row = session.query(TrackRating).\
                         filter(TrackRating.track_id == song_row.id,
                                TrackRating.user_id == user.id).one_or_none()
    if rating_row and rating_row.starred:
        song.set("starred", rating_row.starred.isoformat())
    return song


def fillPlayList(session, row):
    playlist = ET.Element("playlist")
    playlist.set("id", "pl-%d" % row.id)
    playlist.set("name", row.name)
    playlist.set("comment", row.comment if row.comment else "")
    playlist.set("owner", row.owner.name)
    playlist.set("public", "true" if row.public else "false")
    playlist.set("created", row.created.isoformat())
    playlist.set("changed", row.changed.isoformat())
    # FIXME: Join/walk the artist/album/track for art
    fillCoverArt(session, row, playlist, "pl")

    count = 0
    duration = 0
    for trow in row.tracks:
        count += 1
        duration += trow.track.time_secs
        playlist.set("songCount", str(count))
        playlist.set("duration", str(duration))

    for urow in row.users:
        auser = ET.Element("allowedUser")
        auser.text = urow.user.name
        playlist.append(auser)

    return playlist


def fillUser(session, row):
    user = ET.Element("user")
    user.set("username", row.name)
    user.set("email", row.email if row.email else "")
    user.set("scrobblingEnabled", "true" if row.scrobbling else "false")
    for role in Roles.subsonic_roles:
        user.set("%sRole" % role,
                 "true" if role in row.roles else "false")
    return user



import os
import sys
import time
from pkg_resources import load_entry_point

from pyramid.config import Configurator

import unsonic
from . import models, auth
from .views import rest, ui
from .config import HereConfig
from .translogger import ColorTransLogger


__requires__ = 'pyramid>=1.4.3'


NAME = "Unsonic"
UI = None
CONFIG_FILE = None
CONFIG = None
SETTINGS = None


def init(global_config, settings, dbinfo=None):
    global CONFIG_FILE, CONFIG, SETTINGS

    # Stash the global config bits
    CONFIG_FILE = global_config["__file__"]
    unsonic.HERE = global_config["here"]
    SETTINGS = settings
    CONFIG = HereConfig(CONFIG_FILE)
    CONFIG.read()

    # Setup models
    models.init(settings, True, db_info=dbinfo)
    models.load()

    install = "/".join(os.path.dirname(__file__).split("/")[:-1])
    global_config["install"] = install

    unsonic.log.resetupLogging(global_config["__file__"], global_config)


def main(global_config, **settings):
    """This function returns a Pyramid WSGI application."""
    # log.setupMash()
    init(global_config, settings)

    # Pyramid framework
    config = Configurator(settings=settings)
    config.add_static_view('static', 'static', cache_max_age=3600,)
    config.add_route('home', '/', factory="unsonic.views.RouteContext")
    config.add_view('unsonic.views.ui.view', route_name='home',
                    permission=auth.Roles.USERS)
    config.scan()

    global NAME, UI
    NAME = CONFIG.get("unsonic", "name")
    UI = CONFIG.get("unsonic", "ui")

    # Init auth
    auth.init(global_config, config)

    # Init ui
    ui.init(global_config, config)

    # Add the rest interfaces
    config.add_route("rest", "/rest", factory="unsonic.views.rest.RouteContext")
    for cmd in rest.commands.values():
        cmd.settings = settings
        config.add_route(cmd.name, "/rest/" + cmd.name,
                         factory="unsonic.views.rest.RouteContext")
        config.add_view(cmd, route_name=cmd.name, permission=auth.Roles.REST)
        # Clementine calls some API's with the wrong caps.. sigh
        name = cmd.name[0].upper() + cmd.name[1:]
        config.add_route(name, "/rest/" + name,
                         factory="unsonic.views.rest.RouteContext")
        config.add_view(cmd, route_name=name, permission=auth.Roles.REST)

    # Log requests
    app = config.make_wsgi_app()
    return ColorTransLogger(app, setup_console_handler=False)


# Wrapper around the pserve script to catch syntax and import errors
def webServe():
    if "--reload" in sys.argv:
        reload = True
    else:
        reload = False

    wait = 3
    while True:
        try:
            sys.exit(load_entry_point('pyramid>=1.4.3', 'console_scripts',
                                      'pserve')())
            break
        except IndentationError as e:
            if reload:
                print("Failed to (re)start unsonic:", e)
                print("Restarting in %d seconds..." % wait)
            else:
                raise
        except SyntaxError as e:
            if reload:
                print("Failed to (re)start unsonic:", e)
                print("Restarting in %d seconds..." % wait)
            else:
                raise
        except ImportError as e:
            if reload:
                print("Failed to (re)start unsonic:", e)
                print("Restarting in %d seconds..." % wait)
            else:
                raise
        time.sleep(wait)



class ModuleRef(object):
    def __init__(self, obj=None):
        self.__setobj__(obj)

    def __setobj__(self, obj):
        self.__dict__["__obj__"] = obj

    def __getattr__(self, name):
        if not self.__obj__:
            raise AttributeError("ModuleRef.__obj__ is None")
        return getattr(self.__obj__, name)

    def __setattr__(self, name, val):
        self.__obj__.__setattr__(name, val)

    def __call__(self, *args, **kwargs):
        return self.__obj__.__call__(*args, **kwargs)



import xml.etree.ElementTree as ET

from . import Command, registerCmd


@registerCmd
class GetPodcasts(Command):
    name = "getPodcasts.view"
    param_defs = {}
    dbsess = True


    def handleReq(self, session):
        podcasts = ET.Element("podcasts")
        return self.makeResp(child=podcasts)



import os

from pyramid.response import Response, FileResponse

from ... import web
from ...auth import Roles


def view(req):
    resp = req.response

    base = os.path.dirname(__file__)[:-8]
    if req.path == "/":
        resp.content_type = "text/html"
        path = "static/index.html"
    else:
        resp.content_type = "text/javascript"
        path = os.path.join("views", req.path[1:])
    fname = os.path.join(base, path)

    repls = {"rest": req.route_url("rest"),
             "home": req.route_url("home"),
             "user": req.authed_user.name.encode("utf-8")}
    body = open(fname).read()
    for key, val in repls.items():
        body = body.replace("%%%s%%" % str(key), str(val))

    resp.charset = "UTF-8"
    resp.text = body
    return resp


def index(request):
    here = os.path.dirname(__file__)
    index = os.path.join(here, 'index.html')
    return FileResponse(index, request=request)


def notfound(request):
    return Response('Not Found', status='404 Not Found')


def init(global_config, config):
    ui = web.UI
    if not ui:
        config.add_notfound_view(notfound)
    else:
        ui = os.path.abspath(ui)
        config.add_static_view('ui', ui, cache_max_age=3600,
                               factory="unsonic.views.RouteContext",
                               permission=Roles.USERS)



import xml.etree.ElementTree as ET

from . import Command, registerCmd, fillPlayList, NoPerm
from ...models import PlayList, getUserByName


@registerCmd
class GetPlayLists(Command):
    name = "getPlaylists.view"
    param_defs = {"username": {}}
    dbsess = True


    def handleReq(self, session):
        if (not self.params["username"] or
                self.req.authed_user.name == self.params["username"]):
            db_user = self.req.authed_user
        elif self.req.authed_user.isAdmin():
            db_user = getUserByName(session, self.params["username"])
        else:
            raise NoPerm("Can not view another user's playlists unless you "
                         "are an admin")

        playlists = ET.Element("playlists")
        for plrow in session.query(PlayList).filter(
                PlayList.user_id == db_user.id).all():
            playlist = fillPlayList(session, plrow)
            playlists.append(playlist)

        return self.makeResp(child=playlists)



import hashlib
import codecs

from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import forget
from pyramid.view import forbidden_view_config

import unsonic
from unsonic import models
from unsonic.models import Session


class User():
    def __init__(self, db_user):
        self.id = db_user.id
        self.name = db_user.name
        self.password = db_user.password
        self.email = db_user.email
        self.scrobble_user = db_user.scrobble_user
        self.scrobble_pass = db_user.scrobble_pass
        self.scrobble_type = db_user.scrobble_type
        self.roles = []
        for role in db_user.roles:
            self.roles.append(role.name)
        self.avatar = db_user.avatar


    def isAdmin(self):
        return Roles.ADMIN in self.roles


    def isUser(self):
        return Roles.USERS in self.roles


    def isRest(self):
        return Roles.REST in self.roles


class Roles(object):
    # Pyramind roles
    #   ADMIN is for both pyramid and subsonic
    ADMIN = "admin"
    USERS = "users"
    REST = "rest"

    # Subsonic roles, includes admin role
    SETTINGS = "settings"
    DOWNLOAD = "download"
    UPLOAD = "upload"
    PLAYLIST = "playlist"
    COVERART = "coverArt"
    COMMENT = "comment"
    PODCAST = "podcast"
    STREAM = "stream"
    JUKEBOX = "jukebox"
    SHARE = "share"
    VIDEOCONVERSION = "videoConversion"

    # Role groups
    subsonic_roles = [ADMIN, SETTINGS, DOWNLOAD, UPLOAD, PLAYLIST, COVERART,
                      COMMENT, PODCAST, STREAM, JUKEBOX, SHARE, VIDEOCONVERSION]
    admin_roles = [ADMIN, USERS, REST, SETTINGS, DOWNLOAD, PLAYLIST, COVERART,
                   COMMENT, PODCAST, STREAM, JUKEBOX, SHARE, VIDEOCONVERSION]
    def_user_roles = [USERS, REST, DOWNLOAD, PLAYLIST, COVERART, COMMENT,
                      PODCAST, STREAM, SHARE]


# Causes auth challenges to be sent back
@forbidden_view_config()
def forbidden_view(request):
    response = HTTPUnauthorized()
    response.headers.update(forget(request))
    return response


# Support the silly auth schemes that Subsonic has
class SubsonicAuth(BasicAuthAuthenticationPolicy):
    def __init__(self, realm='Realm', debug=False):
        super().__init__(self.authCheck, realm=realm, debug=debug)

    def unauthenticated_userid(self, request):
        if request.headers.get('Authorization'):
            return super().unauthenticated_userid(request)
        elif ("u" not in list(request.params.keys()) or
              ("p" not in list(request.params.keys()) and
               "t" not in list(request.params.keys()) and
               "s" not in list(request.params.keys()))):
            return None
        else:
            return request.params["u"]

    def callback(self, username, request):
        if request.headers.get('Authorization'):
            return super().callback(username, request)
        elif ("u" not in list(request.params.keys()) or
              ("p" not in list(request.params.keys()) and
               "t" not in list(request.params.keys()) and
               "s" not in list(request.params.keys()))):
            return None
        else:
            if "p" in list(request.params.keys()):
                password = request.params["p"]
            else:
                password = None
            return self.check(request.params["u"], password, request)

    # Shared between both auth schemes
    def authCheck(self, username, password, req):
        with Session() as session:
            user = models.getUserByName(session, username)
            if not user or user and not user.password:
                return
            if ("t" in list(req.params.keys()) and
                  "s" in list(req.params.keys())):
                sum = hashlib.md5()
                sum.update(user.password.encode("utf-8"))
                sum.update(req.params["s"].encode("utf-8"))
                if sum.hexdigest() == req.params["t"]:
                    # Stash the user for easy access
                    req.authed_user = user
                    return req.authed_user.roles
                else:
                    return
            elif password.startswith("enc:"):
                try:
                    decode_hex = codecs.getdecoder("hex_codec")
                    password = decode_hex(password[4:])[0]
                    password = password.decode("utf-8")
                except:
                    return
            if user and password == user.password:
                # Stash the user for easy access
                req.authed_user = user
                return req.authed_user.roles


def init(global_config, config):
    authn_policy = SubsonicAuth(realm=unsonic.web.NAME)
    authz_policy = ACLAuthorizationPolicy()
    config.set_authentication_policy(authn_policy)
    config.set_authorization_policy(authz_policy)



from . import Command, registerCmd, InternalError, NoPerm
from ...models import User


@registerCmd
class ChangePassword(Command):
    name = "changePassword.view"
    param_defs = {
        "username": {},
        "password": {"required": True},
        }
    dbsess = True


    def handleReq(self, session):
        name = self.params["username"]
        if not name:
            name = self.req.authed_user.name
        if ((name != self.req.authed_user.name) and
                not self.req.authed_user.isAdmin()):
            raise NoPerm("Can not change a user's password unless you "
                         "are an admin")

        user = session.query(User).filter(User.name == name).one_or_none()
        if not user:
            raise InternalError("User '%s' does not exist" % name)

        user.password = self.params["password"]
        session.add(user)

        return self.makeResp()



# flake8: noqa: F401

# This needs to be very early in the import sequence so the monkeypatch
# works right.
from . import config

from . import models, log, auth


# Used by pkg_resources and Pyramid to load the webapp
from .web import main, webServe


import os
HERE = os.getcwd()
del os



import xml.etree.ElementTree as ET

from . import Command, registerCmd, NoPerm, fillUser
from ...models import User
from ... import auth


@registerCmd
class GetUsers(Command):
    name = "getUsers.view"
    param_defs = {}
    dbsess = True


    def handleReq(self, session):
        if not self.req.authed_user.isAdmin():
            raise NoPerm("Can not get user list unless you are an admin")

        users = ET.Element("users")
        for db_user in session.query(User):
            users.append(fillUser(session, auth.User(db_user)))

        return self.makeResp(child=users)



from .getsimilarsongs import GetSimilarSongs
from . import registerCmd


# TODO: Actually implement
@registerCmd
class GetSimilarSongs2(GetSimilarSongs):
    name = "getSimilarSongs2.view"
    param_defs = GetSimilarSongs.param_defs
    dbsess = True



from . import Command, register
from ..models import listUsers


@register
class ListUsers(Command):
    NAME = "listusers"
    HELP = "List users in the database"


    def _run(self, args=None):
        super()._run()

        args = args or self.args

        print("Users:")
        for uname, roles in listUsers(self.db_session):
            print("   %s:  roles: %s" % (uname, ", ".join(roles)))
        return 0



import xml.etree.ElementTree as ET

from . import Command, registerCmd
from ...models import Tag


@registerCmd
class GetGenres(Command):
    name = "getGenres.view"
    param_defs = {}
    dbsess = True


    def handleReq(self, session):
        genres = ET.Element("genres")
        for row in session.query(Tag).order_by(Tag.name).all():
            genre = ET.Element("genre")
            genre.text = row.name.capitalize()
            # TODO: Make this more better sqlalchemy magic?
            genre.set("songCount",
                      str(session.connection().execute(
                          "SELECT COUNT(*) FROM track_tags WHERE tag_id = %d"
                          % row.id).fetchall()[0][0]))
            # TODO: Revist this when album_tags is populated by mishmash
            genre.set("albumCount",
                      str(session.connection().execute(
                          "SELECT COUNT(*) FROM album_tags WHERE tag_id = %d"
                          % row.id).fetchall()[0][0]))
            genres.append(genre)
        return self.makeResp(child=genres)



import datetime

from . import Command, registerCmd, positive_t, track_t
from ...models import User, PlayQueue


@registerCmd
class SavePlayQueue(Command):
    name = "savePlayQueue.view"
    param_defs = {
        "id": {"type": track_t, "multi": True},
        "current": {"type": track_t},
        "position": {"type": positive_t},
        }
    dbsess = True


    def handleReq(self, session):
        user = (session.query(User).
                filter(User.name == self.req.authed_user.name).one_or_none())
        # Delete the old playqueue
        session.query(PlayQueue).filter(PlayQueue.user_id == user.id).delete()
        # Update the user parts
        user.playqueue_cur = self.params["current"]
        if self.params["position"]:
            user.playqueue_pos = self.params["position"]
        else:
            user.playqueue_pos = 0
        if len(self.params["id"]):
            user.playqueue_mtime = datetime.datetime.now()
        else:
            user.playqueue_mtime = None
        user.playqueue_mby = self.req.user_agent
        session.add(user)
        # Create the playqueue entries
        for track_id in self.params["id"]:
            session.add(PlayQueue(user_id=user.id, track_id=track_id))
        # If no id's, then there is no playqueue and its already been deleted
        return self.makeResp()



from sqlalchemy.orm.exc import NoResultFound

from . import Command, registerCmd, bool_t, track_t, NotFound
from ...models import PlayCount, Track, dbinfo
from ...models import Scrobble as DBScrobble


@registerCmd
class Scrobble(Command):
    name = "scrobble.view"
    param_defs = {
        "id": {"required": True, "type": track_t},
        "time": {},
        "submission": {"type": bool_t, "default": True},
        }
    dbsess = True


    def handleReq(self, session):
        if not self.params["submission"]:
            # User is listening, not scrobbling yet
            dbinfo.users[self.req.authed_user.name].listening = self.params["id"]
        else:
            # Check track id
            try:
                session.query(Track).filter(Track.id == self.params["id"]).one()
            except NoResultFound:
                raise NotFound("Track not found")

            # Inc play count
            try:
                pcount = session.query(PlayCount). \
                           filter(PlayCount.track_id == self.params["id"],
                                  PlayCount.user_id ==
                                  self.req.authed_user.id).one()
                pcount.count = pcount.count + 1
            except NoResultFound:
                pcount = PlayCount(track_id=self.params["id"],
                                   user_id=self.req.authed_user.id,
                                   count=1)
                session.add(pcount)
            session.flush()

            # Local scrobble
            scrobble = DBScrobble(user_id=self.req.authed_user.id,
                                  track_id=self.params["id"])
            session.add(scrobble)
        return self.makeResp()



from pyramid.security import Allow, Authenticated, DENY_ALL

from ..auth import Roles


class RouteContext(object):
    __acl__ = [(Allow, Authenticated, Roles.USERS), DENY_ALL]

    def __init__(self, request):
        pass



# flake8: noqa: F401

from nicfit.command import register
from mishmash import core


class Command(core.Command):
    def _run(self, args=None):
        initAlembic(self.config.get("mishmash", "sqlalchemy.url"))


from ..models import initAlembic

from .adduser import AddUser
from .config import Config
from .deluser import DelUser
from .listusers import ListUsers
from .password import Password
from .serve import Serve



from . import Command, register
from ..models import setUserPassword


@register
class Password(Command):
    NAME = "password"
    HELP = "Change a users password."


    def _initArgParser(self, parser):
        parser.add_argument("username", nargs=1, help="users name")
        parser.add_argument("password", nargs=1, help="users password")


    def _run(self, args=None):
        super()._run()
        args = args or self.args

        if setUserPassword(self.db_session, args.username[0],
                           args.password[0]):
            print("Password set for '%s'." % args.username[0])
            return 0
        else:
            print("User '%s' not found." % args.username[0])
            return -1



import os
import datetime
from argparse import Namespace
from contextlib import contextmanager
from pathlib import Path

from alembic import command
from alembic.config import Config as AlemConfig

import sqlalchemy
from sqlalchemy import (Table, Column, Integer, Text, ForeignKey, String, Enum,
                        DateTime, Index, Boolean, UniqueConstraint, Sequence)
from sqlalchemy.engine import Engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relation
from sqlalchemy.orm.exc import NoResultFound

import mishmash.orm
from mishmash.orm import Base, Artist, Album, Meta, Track    # noqa: F401
from mishmash.orm import Image, Tag, Library, artist_images  # noqa: F401
from mishmash.orm import album_images, track_tags            # noqa: F401
from mishmash.database import init as dbinit


db_url = None
db_engine = None
session_maker = None
dbinfo = Namespace()
ALL = object()


@sqlalchemy.event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    """Allows foreign keys to work in sqlite."""
    import sqlite3
    if dbapi_connection.__class__ is sqlite3.Connection:
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


@contextmanager
def Session():
    """Provide a transactional scope around a series of operations."""
    session = session_maker()
    try:
        yield session
        if hasattr(session, "failed"):
            session.rollback()
        else:
            session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


# DB models
class OrmObject(mishmash.orm.OrmObject):
    @staticmethod
    def loadTable(session):
        pass


class DBInfo(Base, OrmObject):
    __tablename__ = 'un_dbinfo'

    version = Column(String(32), nullable=False, primary_key=True)
    last_sync = Column(DateTime)


    @staticmethod
    def loadTable(session):
        dbinfo.version = None
        dbinfo.old_versions = []
        for dbi in session.query(DBInfo).order_by(DBInfo.version).all():
            if dbinfo.version is None:
                dbinfo.version = dbi.version
            else:
                dbinfo.old_versions.append(dbi.version)


class Config(Base, OrmObject):
    __tablename__ = 'un_config'

    key = Column(String, nullable=False, primary_key=True)
    value = Column(String, nullable=False)
    modified = sqlalchemy.Column(sqlalchemy.DateTime(), nullable=False,
                                 default=datetime.datetime.now)


    @staticmethod
    def loadTable(session):
        pass


class UserConfig(Base, OrmObject):
    __tablename__ = 'un_userconfig'

    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False, primary_key=True)
    key = Column(String, nullable=False, primary_key=True)
    value = Column(String, nullable=False)
    modified = sqlalchemy.Column(sqlalchemy.DateTime(), nullable=False,
                                 default=datetime.datetime.now)


    @staticmethod
    def loadTable(session):
        pass


class User(Base, OrmObject):
    __tablename__ = 'un_users'

    SCROBBLE_TYPES = ["NONE", "LAST_FM", "LIBRE_FM"]
    _scrobble_types = Enum(*SCROBBLE_TYPES, name="scrobble_types")

    # Columns
    id = Column(Integer, Sequence("un_users_id_seq"), primary_key=True)
    name = Column(Text, unique=True)
    password = Column(Text)
    email = Column(Text)
    maxbitrate = Column(Integer, default=0, nullable=False)
    scrobble_user = Column(Text)
    scrobble_pass = Column(Text)
    scrobble_type = Column(_scrobble_types, nullable=False, default="NONE")
    playqueue_cur = Column(Integer, ForeignKey("tracks.id"))
    playqueue_pos = Column(Integer, default=0, nullable=False)
    playqueue_mtime = Column(sqlalchemy.DateTime())
    playqueue_mby = Column(Text)
    avatar = Column(Integer, ForeignKey("images.id"))

    # Relations
    roles = relation("Role", cascade="all, delete-orphan",
                      passive_deletes=True)
    playqueue = relation("PlayQueue", cascade="all, delete-orphan",
                         passive_deletes=True)
    playlists = relation("PlayList", cascade="all, delete-orphan",
                         passive_deletes=True)
    playcounts = relation("PlayCount", cascade="all, delete-orphan",
                          passive_deletes=True)
    scrobbles = relation("Scrobble", cascade="all, delete-orphan",
                          passive_deletes=True)


    @staticmethod
    def loadTable(session):
        dbinfo.users = {}
        for user in session.query(User).all():
            u = Namespace()
            u.name = user.name
            u.listening = None
            dbinfo.users[u.name] = u


class Role(Base, OrmObject):
    __tablename__ = 'un_roles'
    __table_args__ = (UniqueConstraint("user_id", "name"), {})

    id = Column(Integer, Sequence("un_roles_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False)
    name = Column(Text, nullable=False)
    user = relation("User")


class PlayQueue(Base, OrmObject):
    __tablename__ = "un_playqueues"

    id = Column(Integer, Sequence("un_playqueues_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False)
    track_id = Column(Integer, ForeignKey("tracks.id", ondelete='CASCADE'),
                      nullable=False)
    user = relation("User")
    track = relation("Track")


playlist_images = Table("un_playlist_images", Base.metadata,
                        Column("playlist_id", Integer,
                               ForeignKey("un_playlists.id",
                                          ondelete='CASCADE')),
                        Column("img_id", Integer,
                               ForeignKey("images.id")))
'''Pivot table 'playlist_images' for mapping a playlist ID to a value in the
`images` table.'''


class PlayList(Base, OrmObject):
    __tablename__ = 'un_playlists'

    id = Column(Integer, Sequence("un_playlists_id_seq"), primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                      nullable=False)
    name = Column(Text)
    comment = Column(Text)
    public = Column(Integer, default=0)
    created = sqlalchemy.Column(sqlalchemy.DateTime(), nullable=False,
                                default=datetime.datetime.now)
    changed = sqlalchemy.Column(sqlalchemy.DateTime(), nullable=False,
                                default=datetime.datetime.now)
    users = relation("PlayListUser", cascade="all, delete-orphan",
                     passive_deletes=True)
    tracks = relation("PlayListTrack", cascade="all, delete-orphan",
                      passive_deletes=True)
    owner = relation("User")
    images = relation("Image", secondary=playlist_images, cascade="all")
    '''one-to-many playlist images.'''


class PlayListUser(Base, OrmObject):
    __tablename__ = 'un_playlistusers'

    id = Column(Integer, Sequence("un_playlistusers_id_seq"), primary_key=True)
    playlist_id = Column(Integer,
                         ForeignKey("un_playlists.id", ondelete='CASCADE'),
                         nullable=False)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False)
    playlist = relation("PlayList")
    user = relation("User")


class PlayListTrack(Base, OrmObject):
    __tablename__ = 'un_playlisttracks'

    id = Column(Integer, Sequence("un_playlisttracks_id_seq"), primary_key=True)
    playlist_id = Column(Integer, ForeignKey("un_playlists.id",
                         ondelete='CASCADE'), nullable=False)
    track_id = Column(Integer, ForeignKey("tracks.id", ondelete='CASCADE'),
                      nullable=False)
    track = relation("Track")
    playlist = relation("PlayList")


# Modify the Track to include a relation to PlayListTrack
Track.playlist = relation("PlayListTrack")


class ArtistRating(Base, OrmObject):
    __tablename__ = "un_artistratings"

    artist_id = Column(Integer, ForeignKey("artists.id", ondelete='CASCADE'),
                       nullable=False, primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False, primary_key=True)
    rating = Column(Integer, default=None, nullable=True)
    pseudo_rating = Column(Boolean(name="pseudo_rating"), default=True,
                           nullable=False)
    starred = Column(DateTime, default=None, nullable=True)
    pseudo_starred = Column(Boolean(name="pseudo_starred"), default=True,
                            nullable=False)
    artist = relation("Artist")
    user = relation("User")


Artist.rating = relation("ArtistRating")


class AlbumRating(Base, OrmObject):
    __tablename__ = "un_albumratings"

    album_id = Column(Integer, ForeignKey("albums.id", ondelete='CASCADE'),
                      nullable=False, primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False, primary_key=True)
    rating = Column(Integer, default=None, nullable=True)
    pseudo_rating = Column(Boolean(name="pseudo_rating"), default=True,
                           nullable=False)
    starred = Column(DateTime, default=None, nullable=True)
    pseudo_starred = Column(Boolean(name="pseudo_starred"), default=True,
                            nullable=False)
    album = relation("Album")
    user = relation("User")


Album.rating = relation("AlbumRating")


class TrackRating(Base, OrmObject):
    __tablename__ = "un_trackratings"

    track_id = Column(Integer, ForeignKey("tracks.id", ondelete='CASCADE'),
                      nullable=False, primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False, primary_key=True)
    rating = Column(Integer, default=None, nullable=True)
    starred = Column(DateTime, default=None, nullable=True)
    track = relation("Track")
    user = relation("User")


Track.rating = relation("TrackRating")


class PlayCount(Base, OrmObject):
    __tablename__ = "un_playcounts"

    track_id = Column(Integer, ForeignKey("tracks.id", ondelete='CASCADE'),
                      nullable=False, primary_key=True)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False, primary_key=True)
    count = Column(Integer, default=1, nullable=False)
    track = relation("Track")
    user = relation("User")


Track.play_count = relation("PlayCount")


class Scrobble(Base, OrmObject):
    __tablename__ = "un_scrobbles"

    id = Column(Integer, Sequence("un_scrobbles_id_seq"), primary_key=True)
    track_id = Column(Integer, ForeignKey("tracks.id", ondelete='CASCADE'),
                      nullable=False)
    user_id = Column(Integer, ForeignKey("un_users.id", ondelete='CASCADE'),
                     nullable=False)
    tstamp = Column(DateTime, default=datetime.datetime.now, nullable=False)
    track = relation("Track")
    user = relation("User")

    @declared_attr
    def __table_args__(cls):
        return (Index("scrobble_user_index", "user_id"), )


Track.scrobbles = relation("Scrobble")


def _dbUrl(config):
    url = (os.environ.get("MISHMASH_DBURL") or
           config.get("mishmash", "sqlalchemy.url"))
    config.set("mishmash", "sqlalchemy.url", url)
    return url


# Utility functions
def init(settings, webapp=False, db_info=None):
    global db_url, db_engine, session_maker

    if db_info:
        db_url = db_info.url
    else:
        db_url = _dbUrl(web.CONFIG)

    settings["sqlalchemy.url"] = db_url
    web.CONFIG.set("mishmash", "sqlalchemy.url", settings["sqlalchemy.url"])

    settings["sqlalchemy.convert_unicode"] = \
        web.CONFIG.get("mishmash", "sqlalchemy.convert_unicode")
    settings["sqlalchemy.encoding"] = web.CONFIG.get("mishmash",
                                                     "sqlalchemy.encoding")

    if not db_info:
        config = Namespace()
        config.db_url = db_url
        config.various_artists_name = web.CONFIG.get("mishmash",
                                                     "various_artists_name")
        db_info = dbinit(config.db_url)
    db_engine = db_info.engine
    session_maker = db_info.SessionMaker

    initAlembic(db_url)


def initAlembic(url):
    # # Save just for creating new schema revisions
    # Base.metadata.create_all(db_engine)

    # Upgrade to head (i.e. this) revision, or no-op if they match
    alembic_d = Path(__file__).parent / "alembic"
    alembic_cfg = AlemConfig(str(alembic_d / "alembic.ini"))
    alembic_cfg.set_main_option("sqlalchemy.url", url)
    command.upgrade(alembic_cfg, "head")


def load():
    with Session() as session:
        for table in UN_TYPES:
            table.loadTable(session)


def asdict(value):
    ret = {}
    for line in [x.strip() for x in value.splitlines()]:
        if len(line):
            key, val = line.split(":")
            ret[key.strip()] = val.strip()
    return ret


# Table utilities

def addUser(session, username, password, roles):
    try:
        user = User(name=username, password=password)
        session.add(user)
        session.flush()
        for name in roles:
            role = Role(name=name, user_id=user.id)
            session.add(role)
        session.flush()
    except IntegrityError:
        session.failed = True
        return "Failed to add user. User already exists."
    return user


def delUser(session, username):
    session.query(User).filter(User.name == username).delete()
    session.flush()


def listUsers(session):
    users = []
    for user in session.query(User).all():
        users.append((user.name, [g.name for g in user.roles]))
    return users


def setUserPassword(session, uname, password):
    user = getUserByName(session, uname, False)
    if not user:
        return False
    session.add(user)
    user.password = password
    session.flush()
    return True


def getUserByName(session, username, wrap=True):
    try:
        urow = session.query(User).filter(User.name == username).one()
        if wrap:
            return auth.User(urow)
        else:
            return urow
    except NoResultFound:
        return None


def getUserByID(session, id):
    try:
        return auth.User(session.query(User).filter(User.id == id).one())
    except NoResultFound:
        return None


def rateItem(session, user_id, item_id, rating=None, starred=None):
    num = int(item_id[3:])
    artist_id = None
    album_id = None
    if item_id.startswith("ar-"):
        row = session.query(ArtistRating).filter(
            ArtistRating.artist_id == num,
            ArtistRating.user_id == user_id).one_or_none()
        if row is None:
            row = ArtistRating(artist_id=num, user_id=user_id)
            session.add(row)
    elif item_id.startswith("al-"):
        row = session.query(AlbumRating).filter(
            AlbumRating.album_id == num,
            AlbumRating.user_id == user_id).one_or_none()
        if row is None:
            row = AlbumRating(album_id=num, user_id=user_id)
            session.add(row)
            session.flush()
        artist_id = row.album.artist_id
    elif item_id.startswith("tr-"):
        row = session.query(TrackRating).filter(
            TrackRating.track_id == num,
            TrackRating.user_id == user_id).one_or_none()
        if row is None:
            row = TrackRating(track_id=num, user_id=user_id)
            session.add(row)
            session.flush()
        album_id = row.track.album_id
    if rating is not None:
        if rating == 0:
            row.rating = None
            row.pseudo_rating = True
        else:
            row.rating = rating
            row.pseudo_rating = False
    if isinstance(starred, datetime.datetime):
        row.starred = starred
        row.pseudo_starred = False
    elif starred is True:
        row.starred = None
        row.pseudo_starred = True
    session.flush()
    if artist_id or album_id:
        updatePseudoRatings(session=session, user_id=user_id, album_id=album_id,
                            artist_id=artist_id)


def updatePseudoRatings(session, user_id=None, album_id=ALL, artist_id=ALL):
    # Update albums first
    albums = None
    if album_id is ALL:
        albums = session.query(Album).all()
    elif album_id is not None:
        albums = session.query(Album).filter(Album.id == album_id).all()

    if albums:
        for album in albums:
            alrating = session.query(AlbumRating).filter(
                AlbumRating.album_id == album.id,
                AlbumRating.user_id == user_id).one_or_none()
            if alrating is None:
                alrating = AlbumRating(album_id=album.id, user_id=user_id)
                session.add(alrating)
            tracks = session.query(Track). \
                         filter(Track.album_id == album.id).all()
            pseudo = 0
            count = 0
            starred = None
            for track in tracks:
                trating = session.query(TrackRating).filter(
                    TrackRating.track_id == track.id,
                    TrackRating.user_id == user_id).one_or_none()
                if trating:
                    if trating.rating is not None:
                        pseudo += trating.rating
                    if trating.starred is not None:
                        if starred is None or starred > trating.starred:
                            # Extract the oldest starred date
                            starred = trating.starred
                # Down rank for unrated tracks
                count += 1
            if count and alrating.pseudo_rating:
                pseudo = round(float(pseudo) / float(count))
                alrating.rating = pseudo
                alrating.pseudo_rating = True
            if starred and alrating.pseudo_starred:
                alrating.starred = starred
                alrating.pseudo_starred = True

    # Update artists last
    artists = None
    if artist_id is ALL:
        artists = session.query(Artist).all()
    elif artist_id is not None:
        artists = session.query(Artist).filter(Artist.id == artist_id).all()

    if artists:
        for artist in artists:
            arrating = session.query(ArtistRating).filter(
                    ArtistRating.artist_id == artist.id,
                    ArtistRating.user_id == user_id).one_or_none()
            if arrating is None:
                arrating = ArtistRating(user_id=user_id, artist_id=artist.id)
                session.add(arrating)
            if arrating.rating is None or not arrating.pseudo_rating:
                continue
            albums = session.query(Album). \
                         filter(Album.artist_id == artist.id).all()
            pseudo = 0
            count = 0
            starred = None
            for album in albums:
                alrating = session.query(AlbumRating).filter(
                    AlbumRating.album_id == album.id,
                    AlbumRating.user_id == user_id).one_or_none()
                if alrating:
                    if alrating.rating is not None:
                        pseudo += alrating.rating
                    if alrating.starred is not None:
                        if starred is None or starred > alrating.starred:
                            # Extract the oldest starred date
                            starred = alrating.starred

                # Down rank for unrated albums
                count += 1
            if count:
                pseudo = round(float(pseudo) / float(count))
                arrating.rating = pseudo
                arrating.pseudo_rating = True
            if starred and arrating.pseudo_starred:
                arrating.starred = starred
                arrating.pseudo_starred = True


from . import auth, web   # noqa: E402


UN_TYPES = [DBInfo, Config, UserConfig, User, Role, PlayQueue, PlayList,
            PlayListUser, PlayListTrack, ArtistRating, AlbumRating, TrackRating,
            PlayCount, Scrobble]



import datetime

from . import Command, registerCmd, playable_id_t, MissingParam
from ...models import rateItem


@registerCmd
class Star(Command):
    name = "star.view"
    param_defs = {
        "id": {"type": playable_id_t},
        "albumId": {"type": playable_id_t},
        "artistId": {"type": playable_id_t},
        }
    dbsess = True


    def handleReq(self, session):
        if self.params["id"]:
            id = self.params["id"]
        elif self.params["albumId"]:
            id = self.params["albumId"]
        elif self.params["artistId"]:
            id = self.params["artistId"]
        else:
            raise MissingParam("Missing a valid id parameter")

        rateItem(session, self.req.authed_user.id, id,
                 starred=datetime.datetime.now())
        return self.makeResp()



import xml.etree.ElementTree as ET

from . import (Command, registerCmd, fillArtistUser, fillAlbumID3, fillTrackUser,
               folder_t)
from ...models import ArtistRating, AlbumRating, TrackRating


@registerCmd
class GetStarred2(Command):
    name = "getStarred2.view"
    param_defs = {"musicFolderId": {"type": folder_t}}
    dbsess = True


    def handleReq(self, session):
        if self.params["musicFolderId"]:
            lib_id = self.params["musicFolderId"]
        else:
            lib_id = None

        starred = ET.Element("starred2")
        # Artists
        for row in session.query(ArtistRating).filter(
                ArtistRating.starred is not None).all():
            if lib_id is None or row.artist.lib_id == lib_id:
                artist = fillArtistUser(session, row.artist, row,
                                        self.req.authed_user)
                starred.append(artist)

        # Albums
        for row in session.query(AlbumRating).filter(
                AlbumRating.starred is not None).all():
            if lib_id is None or row.album.lib_id == lib_id:
                album = fillAlbumID3(session, row.album, self.req.authed_user,
                                     False)
                starred.append(album)

        # Tracks
        for row in session.query(TrackRating).filter(
                TrackRating.starred is not None).all():
            if lib_id is None or row.track.lib_id == lib_id:
                album = fillTrackUser(session, row.track, row,
                                      self.req.authed_user)
                starred.append(album)

        return self.makeResp(child=starred)



from . import registerCmd
from .createuser import CreateUser


@registerCmd
class UpdateUser(CreateUser):
    name = "updateUser.view"
    param_defs = CreateUser.param_defs
    dbsess = True


    def __init__(self, route, req):
        super().__init__(route, req)
        super().setUpdate(True)



import xml.etree.ElementTree as ET

from . import Command, registerCmd, int_t


# TODO: Actually implement
@registerCmd
class GetTopSongs(Command):
    name = "getTopSongs.view"
    param_defs = {
        "artist": {"required": True},
        "count": {"default": 50, "type": int_t},
        }
    dbsess = True


    # Actually do this for realz once last.fm stuff is hooked up
    def handleReq(self, session):
        songs = ET.Element("topSongs")
        return self.makeResp(child=songs)



from . import (Command, registerCmd, fillPlayList, fillTrack, playlist_t,
               MissingParam)
from ...models import PlayList


@registerCmd
class GetPlayList(Command):
    name = "getPlaylist.view"
    param_defs = {"id": {"required": True, "type": playlist_t}}
    dbsess = True


    def handleReq(self, session):
        plrow = session.query(PlayList). \
                    filter(PlayList.id == self.params["id"]). \
                    filter(PlayList.user_id == self.req.authed_user.id). \
                    one_or_none()
        if not plrow:
            raise MissingParam("Invalid playlist id: %s" % self.params["id"])

        playlist = fillPlayList(session, plrow)
        for pltrack in plrow.tracks:
            entry = fillTrack(session, pltrack.track, name="entry")
            playlist.append(entry)

        return self.makeResp(child=playlist)



"""Initial schema

Revision ID: 7d7ec3901142
Revises:
Create Date: 2017-04-08 16:03:25.287910

"""
from alembic import op
import sqlalchemy as sa
from unsonic.auth import Roles


# revision identifiers, used by Alembic.
revision = '7d7ec3901142'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():

    # DBInfo
    op.create_table(
        'un_dbinfo',
        sa.Column('version', sa.String(length=32), nullable=False),
        sa.Column('last_sync', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('version', name=op.f('pk_un_dbinfo'))
        )

    # Users
    users_t = op.create_table(
        'un_users',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.Text(), nullable=True),
        sa.Column('password', sa.Text(), nullable=True),
        sa.Column('email', sa.Text(), nullable=True),
        sa.Column('maxbitrate', sa.Integer(), nullable=False),
        sa.Column('scrobble_user', sa.Text(), nullable=True),
        sa.Column('scrobble_pass', sa.Text(), nullable=True),
        sa.Column('scrobble_type',
                  sa.Enum('NONE', 'LAST_FM', 'LIBRE_FM', name='scrobble_types'),
                  nullable=False),
        sa.Column('playqueue_cur', sa.Integer(), nullable=True),
        sa.Column('playqueue_pos', sa.Integer(), nullable=False),
        sa.Column('playqueue_mtime', sa.DateTime(), nullable=True),
        sa.Column('playqueue_mby', sa.Text(), nullable=True),
        sa.Column('avatar', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['avatar'], ['images.id'],
                                name=op.f('fk_un_users_avatar_images')),
        sa.ForeignKeyConstraint(['playqueue_cur'], ['tracks.id'],
                                name=op.f('fk_un_users_playqueue_cur_tracks')),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_users')),
        sa.UniqueConstraint('name', name=op.f('uq_un_users_name'))
        )

    # AlbumRatings
    op.create_table(
        'un_albumratings',
        sa.Column('album_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('rating', sa.Integer(), nullable=True),
        sa.Column('pseudo_rating', sa.Boolean(name='pseudo_rating'),
                  nullable=False),
        sa.Column('starred', sa.DateTime(), nullable=True),
        sa.Column('pseudo_starred', sa.Boolean(name='pseudo_starred'),
                  nullable=False),
        sa.ForeignKeyConstraint(['album_id'], ['albums.id'],
                                name=op.f('fk_un_albumratings_album_id_albums'),
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_albumratings_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('album_id', 'user_id',
                                name=op.f('pk_un_albumratings'))
    )

    # ArtistRatings
    op.create_table(
        'un_artistratings',
        sa.Column('artist_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('rating', sa.Integer(), nullable=True),
        sa.Column('pseudo_rating', sa.Boolean(name='pseudo_rating'),
                  nullable=False),
        sa.Column('starred', sa.DateTime(), nullable=True),
        sa.Column('pseudo_starred', sa.Boolean(name='pseudo_starred'),
                  nullable=False),
        sa.ForeignKeyConstraint(
            ['artist_id'], ['artists.id'],
            name=op.f('fk_un_artistratings_artist_id_artists'),
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['user_id'], ['un_users.id'],
            name=op.f('fk_un_artistratings_user_id_un_users'),
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('artist_id', 'user_id',
                                name=op.f('pk_un_artistratings'))
    )

    # Config
    op.create_table(
        'un_config',
        sa.Column('key', sa.String(), nullable=False),
        sa.Column('value', sa.String(), nullable=False),
        sa.Column('modified', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('key', name=op.f('pk_un_config'))
    )

    # User Config
    op.create_table(
        'un_userconfig',
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('key', sa.String(), nullable=False),
        sa.Column('value', sa.String(), nullable=False),
        sa.Column('modified', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_userconfig_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('user_id', 'key', name=op.f('pk_un_userconfig'))
    )

    # PlayCounts
    op.create_table(
        'un_playcounts',
        sa.Column('track_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('count', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['track_id'], ['tracks.id'],
                                name=op.f('fk_un_playcounts_track_id_tracks'),
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_playcounts_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('track_id', 'user_id',
                                name=op.f('pk_un_playcounts'))
    )

    # PlayLists
    op.create_table(
        'un_playlists',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.Text(), nullable=True),
        sa.Column('comment', sa.Text(), nullable=True),
        sa.Column('public', sa.Integer(), nullable=True),
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('changed', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_playlists_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_playlists'))
    )

    # PlayQueues
    op.create_table(
        'un_playqueues',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('track_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['track_id'], ['tracks.id'],
                                name=op.f('fk_un_playqueues_track_id_tracks'),
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_playqueues_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_playqueues'))
    )

    # Roles
    roles_t = op.create_table(
        'un_roles',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_roles_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_roles')),
        sa.UniqueConstraint('user_id', 'name', name=op.f('uq_un_roles_user_id'))
    )

    # Scrobbles
    op.create_table(
        'un_scrobbles',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('track_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('tstamp', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['track_id'], ['tracks.id'],
                                name=op.f('fk_un_scrobbles_track_id_tracks'),
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_scrobbles_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_scrobbles'))
    )
    op.create_index('scrobble_user_index', 'un_scrobbles', ['user_id'],
                    unique=False)

    # TrackRatings
    op.create_table(
        'un_trackratings',
        sa.Column('track_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('rating', sa.Integer(), nullable=True),
        sa.Column('starred', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['track_id'], ['tracks.id'],
                                name=op.f('fk_un_trackratings_track_id_tracks'),
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['user_id'], ['un_users.id'],
                                name=op.f('fk_un_trackratings_user_id_un_users'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('track_id', 'user_id',
                                name=op.f('pk_un_trackratings'))
    )

    # playlist_images
    op.create_table(
        'un_playlist_images',
        sa.Column('playlist_id', sa.Integer(), nullable=True),
        sa.Column('img_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['img_id'], ['images.id'],
            name=op.f('fk_un_playlist_images_img_id_images')),
        sa.ForeignKeyConstraint(
            ['playlist_id'], ['un_playlists.id'],
            name=op.f('fk_un_playlist_images_playlist_id_un_playlists'),
            ondelete='CASCADE')
    )

    # PlayListTracks
    op.create_table(
        'un_playlisttracks',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('playlist_id', sa.Integer(), nullable=False),
        sa.Column('track_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['playlist_id'], ['un_playlists.id'],
            name=op.f('fk_un_playlisttracks_playlist_id_un_playlists'),
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['track_id'], ['tracks.id'],
            name=op.f('fk_un_playlisttracks_track_id_tracks'),
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_playlisttracks'))
    )

    # PlayListUsers
    op.create_table(
        'un_playlistusers',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('playlist_id', sa.Integer(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['playlist_id'], ['un_playlists.id'],
            name=op.f('fk_un_playlistusers_playlist_id_un_playlists'),
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['user_id'], ['un_users.id'],
            name=op.f('fk_un_playlistusers_user_id_un_users'),
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_un_playlistusers'))
    )

    # Prepopulate some data
    op.bulk_insert(users_t, [{"name": "admin", "maxbitrate": 0,
                              "scrobbling": True, "scrobble_type": "NONE",
                              "playqueue_pos": 0}])
    for role in Roles.admin_roles:
        op.bulk_insert(roles_t, [{"user_id": 1, "name": role}])


def downgrade():
    op.drop_table('un_playlistusers')
    op.drop_table('un_playlisttracks')
    op.drop_table('un_playlist_images')
    op.drop_table('un_trackratings')
    op.drop_index('scrobble_user_index', table_name='un_scrobbles')
    op.drop_table('un_scrobbles')
    op.drop_table('un_roles')
    op.drop_table('un_playqueues')
    op.drop_table('un_playlists')
    op.drop_table('un_playcounts')
    op.drop_table('un_config')
    op.drop_table('un_artistratings')
    op.drop_table('un_albumratings')
    op.drop_table('un_users')
    op.drop_table('un_dbinfo')



from . import Command, registerCmd, NoPerm
from ...models import User


@registerCmd
class DeleteUser(Command):
    name = "deleteUser.view"
    param_defs = {
        "username": {"required": True},
        }
    dbsess = True


    def handleReq(self, session):
        name = self.params["username"]
        if ((name != self.req.authed_user.name) and
                not self.req.authed_user.isAdmin()):
            raise NoPerm("Can not delete a user unless you are an admin")

        session.query(User).filter(User.name == name).delete()

        return self.makeResp()



import xml.etree.ElementTree as ET

from . import Command, registerCmd
from ...models import Library


@registerCmd
class GetMusicFolders(Command):
    name = "getMusicFolders.view"
    param_defs = {}
    dbsess = True


    def handleReq(self, session):
        folders = ET.Element("musicFolders")
        for row in session.query(Library).all():
            if row.name == "__null_lib__":
                continue
            folder = ET.Element("musicFolder")
            folders.append(folder)
            folder.set("id", "fl-%s" % row.id)
            folder.set("name", row.name)
        return self.makeResp(child=folders)



import inspect
import unsonic
import unsonic.commands         # noqa: F401
from nicfit import Command

import mishmash
import mishmash.__main__


def buildApp():
    if "web" in Command._all_commands:
        del Command._all_commands["web"]
    if mishmash.commands.web.Web.name in Command._all_commands:
        del Command._all_commands[mishmash.commands.web.Web.name]

    # Temp work around while waiting for mishmash release
    argspec = inspect.getargspec(mishmash.__main__.MishMash.__init__)
    if "progname" in argspec.args:
        return mishmash.__main__.MishMash(progname="unsonic")
    else:
        return mishmash.__main__.MishMash()


def run(args=None):
    app = buildApp()
    return app.run(args)


def main(args=None):
    app = buildApp()
    return app.main(args)


if __name__ == "__main__":
    run()



import xml.etree.ElementTree as ET

from sqlalchemy.sql.expression import func as dbfunc

from eyed3.core import Date as Eyed3Date

from . import (Command, registerCmd, MissingParam, fillAlbumUser, folder_t,
               str_t, int_t)
from ...models import (Artist, Album, AlbumRating, PlayCount, Track,
                       Scrobble, Tag, track_tags)


@registerCmd
class GetAlbumList(Command):
    name = "getAlbumList.view"
    param_defs = {
        "size": {"default": 10, "type": int_t},
        "offset": {"default": 0, "type": int_t},
        "fromYear": {"type": int_t},
        "toYear": {"type": int_t},
        "genre": {"type": str_t},
        "type": {"required": True,
                 "values": ["alphabeticalByName", "alphabeticalByArtist",
                            "frequent", "highest", "newest", "random",
                            "recent", "starred", "byYear", "byGenre"]},
        "musicFolderId": {"type": folder_t},
        }
    dbsess = True

    def __init__(self, route, req, session=None):
        super().__init__(route, req, session)
        self.setParams()


    def setParams(self, list_param="albumList"):
        self.list_param = list_param


    def processRows(self, session, alist, result):
        for row in result:
            album = fillAlbumUser(session, row, None, self.req.authed_user)
            alist.append(album)
            if row.artist:
                album.set("parent", "ar-%d" % row.artist.id)
            else:
                album.set("parent", "UNKNOWN")


    def queryAlbum(self, session):
        if self.params["musicFolderId"]:
            return (session.query(Album).
                    filter(Album.lib_id == self.params["musicFolderId"]))
        else:
            return session.query(Album)


    def handleReq(self, session):
        if self.params["musicFolderId"]:
            lib_id = self.params["musicFolderId"]
        else:
            lib_id = None
        alist = ET.Element(self.list_param)
        size = self.params["size"]
        offset = self.params["offset"]
        limit = offset + size
        if self.params["type"] == "random":
            result = self.queryAlbum(session).\
                         order_by(dbfunc.random()).\
                         offset(offset).\
                         limit(limit)
            self.processRows(session, alist, result)
        elif self.params["type"] == "newest":
            from sqlalchemy import desc as desc_
            albums = session.query(Album)\
                            .order_by(desc_("date_added"))\
                            .offset(offset).limit(limit).all()
            # FIXME: handle new singles here, make sure this works for various
            self.processRows(session, alist, albums)
        elif self.params["type"] == "highest":
            result = session.query(AlbumRating). \
                         filter(AlbumRating.user_id ==
                                self.req.authed_user.id). \
                         order_by(AlbumRating.rating). \
                         offset(offset). \
                         limit(limit)
            albums = []
            for arate in result:
                # TODO: Replace this with the correct sqlalchemy magic
                if lib_id is None or arate.album.lib_id == lib_id:
                    albums.append(arate.album)
            self.processRows(session, alist, albums)
        elif self.params["type"] == "frequent":
            pcounts = session.query(PlayCount).\
                         join(Track).\
                         filter(PlayCount.user_id ==
                                self.req.authed_user.id).\
                         order_by(PlayCount.count).\
                         offset(offset).\
                         limit(limit)
            albums = []
            for pcount in pcounts:
                if not pcount.track.album:
                    continue
                # TODO: Replace this with the correct sqlalchemy magic
                if lib_id and pcount.track.lib_id != lib_id:
                    continue
                if pcount.track.album not in albums:
                    albums.append(pcount.track.album)
                if len(albums) >= size:
                    break
            self.processRows(session, alist, albums)
        elif self.params["type"] == "recent":
            result = session.query(Scrobble).\
                        filter(Scrobble.user_id ==
                               self.req.authed_user.id).\
                        order_by(Scrobble.tstamp.desc()).\
                        offset(offset).\
                        limit(limit)
            albums = []
            for scrobble in result:
                if not scrobble.track.album:
                    continue
                # TODO: Replace this with the correct sqlalchemy magic
                if lib_id and scrobble.track.lib_id != lib_id:
                    continue
                if scrobble.track.album not in albums:
                    albums.append(scrobble.track.album)
                if len(albums) >= size:
                    break
            self.processRows(session, alist, albums)
        elif self.params["type"] == "starred":
            result = session.query(AlbumRating).\
                         filter(AlbumRating.user_id ==
                                self.req.authed_user.id).\
                         filter(AlbumRating.starred is not None).\
                         order_by(AlbumRating.starred).\
                         offset(offset).\
                         limit(limit)
            albums = []
            for arate in result:
                # TODO: Replace this with the correct sqlalchemy magic
                if lib_id is None or arate.album.lib_id != lib_id:
                    albums.append(arate.album)
            self.processRows(session, alist, albums)
        elif self.params["type"] == "alphabeticalByName":
            result = self.queryAlbum(session).\
                         order_by(Album.title).\
                         offset(offset).\
                         limit(limit)
            self.processRows(session, alist, result)
        elif self.params["type"] == "alphabeticalByArtist":
            artists = session.query(Artist).\
                          order_by("name").\
                          limit(limit)
            for artist in artists:
                albums = self.queryAlbum(session).\
                             filter(Album.artist_id == artist.id).\
                             order_by(Album.title).\
                             offset(offset).\
                             limit(limit)
                self.processRows(session, alist, artist.albums)
        elif self.params["type"] == "byYear":
            to_year = self.params["toYear"]
            from_year = self.params["fromYear"]
            if not to_year or not from_year:
                raise MissingParam("Missing proper toYear/fromYear parameters "
                                   "for search byYear")
            if from_year > to_year:
                desc = True
                to_year = Eyed3Date(year=from_year, month=12, day=31)
                from_year = Eyed3Date(year=to_year, month=1, day=1)
            else:
                desc = False
                from_year = Eyed3Date(year=from_year, month=1, day=1)
                to_year = Eyed3Date(year=to_year, month=12, day=31)
            results = set()
            second = False
            for date_row in [Album.original_release_date, Album.release_date,
                             Album.recording_date]:
                res = self.queryAlbum(session).\
                          filter(date_row >= from_year, date_row <= to_year).\
                          offset(offset).\
                          limit(limit)
                if not second:
                    second = True
                    for row in res.all():
                        results.add(row)
                else:
                    for row in res.all():
                        # If no orig date, go ahead and add it
                        if not row.original_release_date:
                            results.add(row)
                        # Add only if the orig date is in the range and ignore
                        # rel date
                        elif ((row.original_release_date >= from_year) and
                              (row.original_release_date <= to_year)):
                            results.add(row)
            results = list(results)
            results.sort(key=lambda x: x.getBestDate(), reverse=desc)
            self.processRows(session, alist, results)
        elif self.params["type"] == "byGenre":
            genre = self.params["genre"]
            if not genre:
                raise MissingParam("Missing genre param when searching byGenre")

            tag = session.query(Tag).\
                filter(dbfunc.lower(Tag.name) == genre.lower()).one_or_none()
            if not tag:
                return self.makeResp(child=alist)

            result = session.query(Album).\
                join(Track).\
                join(track_tags).\
                filter(Album.id == Track.album_id,
                       Track.id == track_tags.c.track_id,
                       track_tags.c.tag_id == tag.id).\
                group_by(Album).\
                order_by(Album.title).\
                offset(offset).\
                limit(limit)
            self.processRows(session, alist, result)
        else:
            raise MissingParam("Unsupported type")

        return self.makeResp(child=alist)



import xml.etree.ElementTree as ET

from . import (Command, registerCmd, positive_t, fillArtist, fillAlbum,
               fillTrack, folder_t)
from ...models import Artist, Album, Track


@registerCmd
class Search2(Command):
    name = "search2.view"
    param_defs = {
        "query": {"required": True},
        "artistCount": {"default": 20, "type": positive_t},
        "artistOffset": {"default": 0, "type": positive_t},
        "albumCount": {"default": 20, "type": positive_t},
        "albumOffset": {"default": 0, "type": positive_t},
        "songCount": {"default": 20, "type": positive_t},
        "songOffset": {"default": 0, "type": positive_t},
        "musicFolderId": {"type": folder_t},
        }
    dbsess = True


    def query(self, session, klass):
        if self.params["musicFolderId"]:
            return (session.query(klass).
                    filter(klass.lib_id == self.params["musicFolderId"]))
        else:
            return session.query(klass)


    def handleReq(self, session):
        query = self.params["query"]
        ar_count = self.params["artistCount"]
        ar_off = self.params["artistOffset"]
        al_count = self.params["albumCount"]
        al_off = self.params["albumOffset"]
        tr_count = self.params["songCount"]
        tr_off = self.params["songOffset"]

        result = ET.Element("searchResult2")
        if ar_count:
            for row in self.query(session, Artist). \
                           filter(Artist.name.ilike("%%%s%%" % query)). \
                           limit(ar_count). \
                           offset(ar_off):
                artist = fillArtist(session, row)
                result.append(artist)
        if al_count:
            for row in self.query(session, Album). \
                           filter(Album.title.ilike("%%%s%%" % query)). \
                           limit(al_count). \
                           offset(al_off):
                album = fillAlbum(session, row)
                result.append(album)
        if tr_count:
            for row in self.query(session, Track). \
                           filter(Track.title.ilike("%%%s%%" % query)). \
                           limit(tr_count). \
                           offset(tr_off):
                track = fillTrack(session, row)
                result.append(track)
        return self.makeResp(child=result)



from . import Command, registerCmd, NoPerm, fillUser
from ...models import getUserByName


@registerCmd
class GetUser(Command):
    name = "getUser.view"
    param_defs = {
        "username": {"required": True},
        }
    dbsess = True


    def handleReq(self, session):
        if self.req.authed_user.name == self.params["username"]:
            db_user = self.req.authed_user
        elif self.req.authed_user.isAdmin():
            db_user = getUserByName(session, self.params["username"])
        else:
            raise NoPerm("Can not view a user other than yourself unless you "
                         "are an admin")

        return self.makeResp(child=fillUser(session, db_user))



from . import registerCmd
from .getmusicdirectory import GetMusicDirectory


@registerCmd
class GetSong(GetMusicDirectory):
    name = "getSong.view"

    def __init__(self, route, req):
        super(GetSong, self).__init__(route, req)
        self.setParams(dir_param="song", album_param="song", track_param="song")



import xml.etree.ElementTree as ET

from . import Command, registerCmd, fillArtist, folder_t
from ...models import Artist, Album


@registerCmd
class GetArtists(Command):
    name = "getArtists.view"
    param_defs = {"musicFolderId": {"type": folder_t}}
    dbsess = True


    def handleReq(self, session):
        artists = ET.Element("artists")
        # TODO: find the actual ignored articles
        artists.set("ignoredArticles", "")
        index_group = None
        q = session.query(Artist)
        if self.params["musicFolderId"]:
            q = q.filter(Artist.lib_id == self.params["musicFolderId"])
        for row in q.order_by(Artist.sort_name).all():
            first = row.sort_name[0].upper()
            if index_group != first:
                index_group = first
                index = ET.Element("index")
                artists.append(index)
                index.set("name", index_group)
            artist = fillArtist(session, row)
            index.append(artist)
        for index in artists:
            for artist in index:
                count = 0
                for album in session.query(Album).filter(
                        Album.artist_id == int(artist.get("id")[3:])).all():
                    count = count + 1
                artist.set("albumCount", str(count))
        return self.makeResp(child=artists)



from . import Command, registerCmd, MissingParam, track_t, playlist_t
from .getplaylist import GetPlayList
from ...models import PlayList, PlayListTrack, Track


@registerCmd
class CreatePlayList(Command):
    name = "createPlaylist.view"
    param_defs = {
        "playlistId": {"type": playlist_t},
        "name": {},
        "songId": {"multi": True, "type": track_t},
        }
    dbsess = True


    def handleReq(self, session):
        plid = self.params["playlistId"]
        if plid and self.params["name"]:
            raise MissingParam(
                "Can't supply playlistId and name at the same time")
        elif (not plid and not self.params["name"]):
            raise MissingParam("Must supply playlistId or name")

        if plid:
            # Update/append case
            plist = session.query(PlayList).filter(
                PlayList.id == plid).one_or_none()
            if plist is None:
                raise MissingParam("Invalid playlistId: pl-%s" % plid)
        else:
            # New creation case
            plist = PlayList(user_id=self.req.authed_user.id,
                             name=self.params["name"])
            session.add(plist)

        # Add tracks
        for sid in self.params["songId"]:
            track = session.query(Track).filter(Track.id == sid).one_or_none()
            if track is None:
                raise MissingParam("Invalid songId: tr-%s" % sid)
            pltrack = PlayListTrack(track_id=track.id, playlist_id=plist.id)
            session.add(pltrack)

        session.flush()

        # Fun little hack to return the newly created playlist
        self.params["id"] = plist.id
        return GetPlayList.handleReq(self, session)



import time

from paste.translogger import TransLogger

from nicfit.console.ansi import Fg


class ColorTransLogger(TransLogger):

    format = ('%(REMOTE_ADDR)s - %(REMOTE_USER)s [%(time)s] '
              '"%(REQUEST_METHOD)s %(REQUEST_URI)s %(HTTP_VERSION)s" '
              '%(status)s %(bytes)s "%(HTTP_REFERER)s" "%(HTTP_USER_AGENT)s"')


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def write_log(self, environ, method, req_uri, start, status, bytes):
        if bytes is None:
            bytes = '-'
        if time.daylight:
                offset = time.altzone / 60 / 60 * -100
        else:
                offset = time.timezone / 60 / 60 * -100
        if offset >= 0:
                offset = "+%0.4d" % (offset)
        elif offset < 0:
                offset = "%0.4d" % (offset)
        remote_addr = '-'
        if environ.get('HTTP_X_FORWARDED_FOR'):
            remote_addr = environ['HTTP_X_FORWARDED_FOR']
        elif environ.get('REMOTE_ADDR'):
            remote_addr = environ['REMOTE_ADDR']
        stat = status.split(None, 1)[0]
        d = {
            'REMOTE_ADDR': remote_addr,
            'REMOTE_USER': environ.get('REMOTE_USER') or '-',
            'REQUEST_METHOD': method,
            'REQUEST_URI': req_uri,
            'HTTP_VERSION': environ.get('SERVER_PROTOCOL'),
            'time': time.strftime('%d/%b/%Y:%H:%M:%S ', start) + offset,
            'status': stat,
            'bytes': bytes,
            'HTTP_REFERER': environ.get('HTTP_REFERER', '-'),
            'HTTP_USER_AGENT': environ.get('HTTP_USER_AGENT', '-'),
            }
        message = self.format % d
        stat = int(stat)
        if stat >= 200 and stat < 300:
            message = Fg.green(message)
        if stat >= 400 and stat < 500:
            message = Fg.yellow(message)
        if stat >= 500:
            message = Fg.red(message)
        self.logger.log(self.logging_level, message)



import time
from datetime import datetime
import xml.etree.ElementTree as ET

from . import Command, registerCmd


@registerCmd
class GetLicense(Command):
    name = "getLicense.view"
    param_defs = {}

    def handleReq(self):
        license = ET.Element("license")
        license.set("valid", "true")
        license.set("email", "foo@bar.com")
        now = datetime.fromtimestamp(time.time() + 31536000)
        license.set("licenseExpires", now.isoformat())
        return self.makeResp(child=license)



import os
import sys
import argparse
import shutil

from . import Command, register


@register
class Serve(Command):
    NAME = "serve"
    HELP = "Run the unsonic web interface."
    DESC = ("Run the unsonic web interface using the Pyramid pserve script. "
            "This interface is HTTP only and can not do HTTPS. Please read the "
            "documentation for instructions on how to run Unsonic with HTTPS.")


    def _initArgParser(self, parser):
        parser.add_argument("pserve_args", nargs=argparse.REMAINDER,
                            help="pyramid pserve arguments. Put a '--' before "
                                 "any pserve arguments.")


    def run(self, args, config):
        pargs = args.pserve_args

        if not config.filename:
            print("No config file specified. Must specify a config file.")
            sys.exit(-1)

        # unsonic-server is needed to keep the cmdline args for pserve for its
        # reload option with hupper, otherwise pserve/hupper gets confused.
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../",
                                            "bin/unsonic-server"))
        if not os.access(path, os.X_OK):
            path = shutil.which("unsonic-server")
        if not path:
            print("Failed to find the unsonic-server command.")
            sys.exit(-1)
        argv = [path]
        if len(pargs) and pargs[0] == "--":
            pargs = pargs[1:]
        argv.extend(pargs)
        argv.append(str(config.filename))
        print(" ".join(argv))
        os.execv(path, argv)



from . import Command, registerCmd, playable_id_t, int_t
from ...models import rateItem


@registerCmd
class SetRating(Command):
    name = "setRating.view"
    param_defs = {
        "id": {"type": playable_id_t, "required": True},
        "rating": {"type": int_t, "required": True},
        }
    dbsess = True


    def handleReq(self, session):
        rateItem(session, self.req.authed_user.id, self.params["id"],
                 rating=self.params["rating"])
        return self.makeResp()



import sys
import logging

try:
    # Location of setup_logging in older pyramid versions
    from pyramid.scripts.common import setup_logging
except:
    from pyramid.paster import setup_logging

import eyed3.utils.console


log = logging.getLogger("unsonic")


class StdoutFormatter(logging.Formatter):
    DEFAULT_FORMAT = '%(message)s'

    def __init__(self):
        logging.Formatter.__init__(self, self.DEFAULT_FORMAT)


class FileWrapper(object):
    """A wrapper class for file and file-like objects to provide the same API."""

    def __init__(self, real_file, file_like):
        self._real_file = real_file
        self._file_like = file_like

    def write(self, *args, **kwargs):
        return self._file_like.info(*args, **kwargs)

    def flush(self):
        return self._file_like.handlers[0].flush()

    def isatty(self):
        return self._real_file.isatty()


def setupMash():
    stdout_l = logging.getLogger("mishmash.stdout")
    # FIXME figure out where to put this log file for normal runs
    stdout_handler = logging.StdoutHandler(open("/tmp/foo", "w+"))
    stdout_handler.setFormatter(StdoutFormatter())
    stdout_l.addHandler(stdout_handler)
    stdout_l.setLevel(logging.INFO)
    stdout_l.propagate = False

    wrapper = FileWrapper(sys.stdout, stdout_l)
    eyed3.utils.console.setOutput(wrapper, wrapper)


def resetupLogging(config_uri, global_config):
    setup_logging(config_uri, global_config)



from . import Command, registerCmd, InternalError, NoPerm, bool_t, bitrate_t
from ...models import User, Role
from ...auth import Roles


@registerCmd
class CreateUser(Command):
    name = "createUser.view"
    param_defs = {
        "username": {"required": True},
        "password": {"required": True},
        "email": {},
        "maxBitRate": {"type": bitrate_t, "default": 0},
        # "ldapAuthenticated"  # Skipping
        "adminRole": {"type": bool_t},
        "settingsRole": {"type": bool_t},
        "streamRole": {"type": bool_t},
        "jukeboxRole": {"type": bool_t},
        "downloadRole": {"type": bool_t},
        "uploadRole": {"type": bool_t},
        "playlistRole": {"type": bool_t},
        "coverArtRole": {"type": bool_t},
        "commentRole": {"type": bool_t},
        "podcastRole": {"type": bool_t},
        "shareRole": {"type": bool_t},
        "videoConversionRole": {"type": bool_t},
        # "musicFolderId": {}, # TODO
        }
    dbsess = True
    role_names = {"adminRole": Roles.ADMIN, "settingsRole": Roles.SETTINGS,
                  "streamRole": Roles.STREAM, "jukeboxRole": Roles.JUKEBOX,
                  "downloadRole": Roles.DOWNLOAD, "uploadRole": Roles.UPLOAD,
                  "playlistRole": Roles.PLAYLIST, "coverArtRole": Roles.COVERART,
                  "commentRole": Roles.COMMENT, "podcastRole": Roles.PODCAST,
                  "shareRole": Roles.SHARE,
                  "videoConversionRole": Roles.VIDEOCONVERSION}


    def __init__(self, route, req, session=None):
        super().__init__(route, req, session=session)
        self.update = False


    def setUpdate(self, val):
        self.update = val


    def handleReq(self, session):
        if not self.req.authed_user.isAdmin():
            raise NoPerm("Can not create a user unless you are an admin")

        name = self.params["username"]
        user = session.query(User).filter(User.name == name).one_or_none()
        if user and not self.update:
            raise InternalError("User '%s' already exists" % name)
        if not user and self.update:
            raise InternalError("User '%s' does not exist" % name)

        if not self.update:
            user = User()
        user.name = name
        user.password = self.params["password"]
        user.email = self.params["email"]
        user.maxbitrate = self.params["maxBitRate"]
        session.add(user)
        session.flush()

        for role, un_role in self.role_names.items():
            if (self.params[role] and
                (session.query(Role).filter(Role.user_id == user.id).
                 filter(Role.name == un_role).one_or_none()) is None):
                session.add(Role(user_id=user.id, name=un_role))
            if self.params[role] is False:
                session.query(Role).filter(Role.user_id == user.id).\
                    filter(Role.name == un_role).delete()

        if not self.update:
            session.add(Role(user_id=user.id, name=Roles.USERS))
            session.add(Role(user_id=user.id, name=Roles.REST))

        return self.makeResp()



from . import (Command, registerCmd, MissingParam, track_t, bool_t, positive_t,
               playlist_t)
from ...models import PlayList, PlayListTrack, Track


@registerCmd
class UpdatePlayList(Command):
    name = "updatePlaylist.view"
    param_defs = {
        "playlistId": {"required": True, "type": playlist_t},
        "name": {},
        "comment": {},
        "public": {"type": bool_t},
        "songIdToAdd": {"multi": True, "type": track_t},
        "songIndexToRemove": {"multi": True, "type": positive_t},
        }
    dbsess = True


    def handleReq(self, session):
        plist = session.query(PlayList).filter(
            PlayList.id == self.params["playlistId"]).one_or_none()
        if plist is None:
            raise MissingParam("Invalid playlistId: pl-%s" %
                               self.params["playlistId"])

        if self.params["name"]:
            plist.name = self.params["name"]
        if self.params["comment"]:
            plist.comment = self.params["comment"]
        if self.params["public"] is not None:
            plist.public = self.params["public"]

        # Delete songs
        if len(self.params["songIndexToRemove"]):
            tracks = session.query(PlayListTrack).filter(
                PlayListTrack.playlist_id == self.params["playlistId"]).all()
            tracks_len = len(tracks)
            for index in self.params["songIndexToRemove"]:
                if index >= tracks_len:
                    raise MissingParam("Invalid songIndexToRemove: %d" % index)
                res = session.query(PlayListTrack). \
                          filter(PlayListTrack.id == tracks[index].id)
                if not res.one_or_none():
                    raise MissingParam("Invalid track id: tr-%s" %
                                       tracks[index].id)
                res.delete()

        # Add songs
        for sid in self.params["songIdToAdd"]:
            track = session.query(Track).filter(Track.id == sid).one()
            if track is None:
                raise MissingParam("Invalid songId: tr-%s" % sid)
            pltrack = PlayListTrack(track_id=track.id, playlist_id=plist.id)
            session.add(pltrack)

        plist.changed = plist.changed.now()

        return self.makeResp()



from . import (Command, registerCmd, NotFound, artist_t, fillAlbumID3,
               fillArtistUser)
from ...models import Artist, Album, Track


@registerCmd
class GetArtist(Command):
    name = "getArtist.view"
    param_defs = {"id": {"required": True, "type": artist_t}}
    dbsess = True


    def handleReq(self, session):
        artist = None
        for row in session.query(Artist).filter(
                Artist.id == self.params["id"]).all():
            artist = fillArtistUser(session, row, None, self.req.authed_user)
        if artist is None:
            raise NotFound(self.req.params["id"])
        album_count = 0
        for row in session.query(Album).filter(
                Album.artist_id == self.params["id"]).all():
            album_count += 1
            album = fillAlbumID3(session, row, self.req.authed_user, False)
            artist.append(album)
        for album in artist:
            if album.tag != "album":
                continue
            song_count = 0
            duration = 0
            for row in session.query(Track).filter(
                    Track.album_id == int(album.get("id")[3:])).all():
                song_count = song_count + 1
                duration = duration + row.time_secs
            album.set("songCount", str(song_count))
            album.set("duration", str(duration))
        artist.set("albumCount", str(album_count))
        return self.makeResp(child=artist)



from . import Command, registerCmd


@registerCmd
class GetChatMessages(Command):
    name = "getChatMessages.view"
    param_defs = {"since": {}}
    dbsess = True


    def handleReq(self, session):
        return self.makeResp()



import os
import runpy

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()
CHANGES = open(os.path.join(here, 'CHANGES.md')).read()

# Extract the version from unsonic
VERSION = runpy.run_path(os.path.join(here, "unsonic/version.py"))["VERSION"]


def requirements(filename):
    if os.path.exists(filename):
        return [l for l in open(filename).read().splitlines()
                    if not l.startswith("#")]
    else:
        return ""


setup(name='unsonic',
      version=VERSION,
      description='Unsonic, the un-Subsonic music server.',
      long_description=(README + '\n\n' + "*" * 30 + " CHANGES " + "*" * 30 +
                        "\n\n" + CHANGES),
      classifiers=[
          "Intended Audience :: End Users/Desktop",
          "Operating System :: POSIX",
          "Natural Language :: English",
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          ],
      author='Chris Newton',
      author_email='redshodan@gmail.com',
      url='https://github.com/redshodan/unsonic',
      keywords=["unsonic", "mishmash", "eyed3", "web", "wsgi", "bfg",
                "pylons", "pyramid"],
      packages=find_packages(),
      python_requires='>=3.6',
      include_package_data=True,
      zip_safe=False,
      platforms=["Any"],
      test_suite='unsonic',
      install_requires=requirements("requirements.txt"),
      setup_requires=['pytest-runner'],
      tests_require=requirements("requirements-test.txt"),
      license="GPLv2",
      entry_points={
          'paste.app_factory': ["main = unsonic:main"],
          'console_scripts': ['unsonic = unsonic.__main__:main',
                              'unsonic-server = unsonic:webServe'],
      },
      )



import xml.etree.ElementTree as ET

from . import Command, registerCmd, int_t, track_t


# TODO: Actually implement
@registerCmd
class GetSimilarSongs(Command):
    name = "getSimilarSongs.view"
    param_defs = {
        "id": {"required": True, "type": track_t},
        "count": {"default": 50, "type": int_t},
        }
    dbsess = True


    # Actually do this for realz once last.fm stuff is hooked up
    def handleReq(self, session):
        ainfo = ET.Element("similarSongs")
        return self.makeResp(child=ainfo)



from . import Command, registerCmd, playlist_t, MissingParam
from ...models import PlayList


@registerCmd
class DeletePlayList(Command):
    name = "deletePlaylist.view"
    param_defs = {"id": {"required": True, "type": playlist_t}}
    dbsess = True


    def handleReq(self, session):
        res = session.query(PlayList). \
                    filter(PlayList.id == self.params["id"]). \
                    filter(PlayList.user_id == self.req.authed_user.id)
        if not res.one_or_none():
            raise MissingParam("Invalid playlist id: %s" % self.params["id"])
        res.delete()

        return self.makeResp()



from . import Command, register
from ..models import User


@register
class DelUser(Command):
    NAME = "deluser"
    HELP = "Delete a user from the database."


    def _initArgParser(self, parser):
        parser.add_argument("username", nargs=1, help="users name")


    def _run(self, args=None):
        super()._run()
        args = args or self.args

        res = self.db_session.query(User).filter(User.name == args.username[0])
        if res.one_or_none():
            res.delete()
            print("Deleted user '%s'." % args.username[0])
            return 0
        else:
            print("User '%s' doesn't exists." % args.username[0])
            return -1



import os
import configparser

# This needs to be very early in the import sequence so the monkeypatch
# works right.
import mishmash
from mishmash.config import Config as MishConfig


CONFIG = None


class HereConfig(MishConfig):
    def __init__(self, filename, **kwargs):
        super().__init__(filename, **kwargs)
        global CONFIG
        CONFIG = self
        # %(install)s
        install = "/".join(os.path.dirname(__file__).split("/")[:-1])
        self.set(configparser.DEFAULTSECT, "install", install)
        if filename:
            # %(here)s
            here = os.path.dirname(os.path.abspath(filename))
            self.set(configparser.DEFAULTSECT, "here", here)
        else:
            self.set(configparser.DEFAULTSECT, "here", install)


    def get(self, section, key, **kwargs):
        val = super().get(section, key, **kwargs)
        if val and "%(here)s" in val:
            return val.replace("%(here)s",
                               super().get(configparser.DEFAULTSECT, "here"))
        elif val and "%(install)s" in val:
            return val.replace("%(install)s",
                               super().get(configparser.DEFAULTSECT, "install"))
        else:
            return val


mishmash.config.Config = HereConfig



import xml.etree.ElementTree as ET

from sqlalchemy import and_
from sqlalchemy.sql.expression import func as dbfunc

from . import Command, registerCmd, fillTrack, year_t, folder_t, int_t, str_t
from ...models import Album, Track, Tag, track_tags


@registerCmd
class GetRandomSongs(Command):
    name = "getRandomSongs.view"
    param_defs = {
        "size": {"default": 10, "type": int_t},
        "genre": {"type": str_t},
        "fromYear": {"type": year_t},
        "toYear": {"type": year_t},
        "musicFolderId": {"type": folder_t},
        }
    dbsess = True


    def query(self, session):
        if self.params["musicFolderId"]:
            return (session.query(Track).
                    filter(Track.lib_id == self.params["musicFolderId"]))
        else:
            return session.query(Track)


    def handleReq(self, session):
        random_songs = ET.Element("randomSongs")
        genre = self.params["genre"]
        fy = self.params["fromYear"]
        ty = self.params["toYear"]
        if genre:
            tag = session.query(Tag).\
                filter(dbfunc.lower(Tag.name) == genre.lower()).one_or_none()
            if not tag:
                return self.makeResp(child=random_songs)
            rows = (self.query(session).join(track_tags).
                    filter(and_(Track.id == track_tags.c.track_id,
                                track_tags.c.tag_id == tag.id)).
                    order_by(dbfunc.random()).limit(self.params["size"]))
        elif fy and ty:
            rows = (self.query(session).
                    join(Album).filter(Album.release_date >= fy and
                                       Album.release_date <= ty).
                    order_by(dbfunc.random()).limit(self.params["size"]))
        elif fy:
            rows = (self.query(session).
                    join(Album).filter(Album.release_date >= fy).
                    order_by(dbfunc.random()).limit(self.params["size"]))
        elif ty:
            rows = (self.query(session).
                    join(Album).filter(Album.release_date <= ty).
                    order_by(dbfunc.random()).limit(self.params["size"]))
        else:
            rows = (self.query(session).
                    order_by(dbfunc.random()).limit(self.params["size"]))
        for row in rows:
            song = fillTrack(session, row)
            random_songs.append(song)
        return self.makeResp(child=random_songs)



import xml.etree.ElementTree as ET

from . import Command, registerCmd, strDate, fillTrackUser
from ...models import User


@registerCmd
class GetPlayQueue(Command):
    name = "getPlayQueue.view"
    param_defs = {}
    dbsess = True


    def handleReq(self, session):
        user = (session.query(User).
                filter(User.name == self.req.authed_user.name).one_or_none())
        if not user.playqueue_mtime:
            return self.makeResp()
        playq = ET.Element("playQueue")
        playq.set("username", user.name)
        if user.playqueue_cur:
            playq.set("current", str(user.playqueue_cur))
        if user.playqueue_pos:
            playq.set("position", str(user.playqueue_pos))
        playq.set("changed", strDate(user.playqueue_mtime))
        playq.set("changedBy", user.playqueue_mby)

        for pq in user.playqueue:
            playq.append(fillTrackUser(session, pq.track, None,
                                       self.req.authed_user, "entry"))

        return self.makeResp(child=playq)



import xml.etree.ElementTree as ET

from sqlalchemy import and_
from sqlalchemy.sql.expression import func as dbfunc

from . import Command, registerCmd, fillTrackUser, folder_t, str_t, int_t
from ...models import Track, Tag, track_tags


@registerCmd
class GetSongsByGenre(Command):
    name = "getSongsByGenre.view"
    param_defs = {
        "genre": {"required": True, "type": str_t},
        "count": {"default": 10, "type": int_t},
        "offset": {"default": 0, "type": int_t},
        "musicFolderId": {"type": folder_t},
        }
    dbsess = True


    def handleReq(self, session):
        songs = ET.Element("songsByGenre")
        genre = self.params["genre"]
        count = self.params["count"]
        offset = self.params["offset"]
        limit = offset + count
        if self.params["musicFolderId"]:
            lib_id = self.params["musicFolderId"]
        else:
            lib_id = None

        # TODO: When mishmash does album tags/genres, do that too
        # TODO: Cache tags?
        tag = session.query(Tag).\
            filter(dbfunc.lower(Tag.name) == genre.lower()).one_or_none()
        if not tag:
            return self.makeResp(child=songs)
        q = session.query(Track).\
            join(track_tags).\
            filter(and_(Track.id == track_tags.c.track_id,
                        track_tags.c.tag_id == tag.id))
        if lib_id:
            q = q.filter(Track.lib_id == lib_id)
        result = q.order_by(Track.title).\
            offset(offset).\
            limit(limit)
        for row in result:
            song = fillTrackUser(session, row, None, self.req.authed_user,
                                 "song")
            songs.append(song)
            if row.album:
                song.set("parent", "al-%d" % row.album.id)
            else:
                song.set("parent", "UNKNOWN")

        return self.makeResp(child=songs)
